/*! \file
    \brief ICP code for JM60 or UF32

    \verbatim
    Opensource BDM  - interface DLL
    Based on TBDML interface DLL by
    Copyright (C) 2005  Daniel Malik

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    \endverbatim

*/
#include <windows.h>
#include <stdio.h>
#ifdef TBLCF
#include "interface.h"
#else
#include "interface.h"
#endif
#include "common_usb.h"
#include "Configure.h"
#include "log.h"
#include "version.h"
#include "Common.h"
#include "ICP.h"
#include "Resource.h"

#include "FlashImage.h" // Contains the BDM Firmware

//! BDM microcontroller types
static const char *BDMCPUTypes[]={"JB16","????","JM16","UF32"};

//! BDM hardware descriptions
static const char *const hardwareDescriptions[] = {
   "Unknown or invalid hardware",                  // 0
   "USBDM - Extended TBDML/OSBDM (JB16DWE)",       // 1
   "TBDML - Minimal TBDML",                        // 2
   "TBDML - Minimal TBDML (Internal version)",     // 3
   "OSBDM - Original OSBDM",                       // 4
   "Witztronics TBDML/OSBDM",                      // 5
   "OSBDM+ - Extended OSBDM (RS08 support)",       // 6
   "USBDM - Extended TBDML/OSBDM (JMxxCLD)",       // 7
   "USBDM - Extended TBDML/OSBDM (JMxxCLC)",       // 8
   "USBSPYDER08 - SofTec USBSPYDER08      ",       // 9
   "USBDM - Extended TBDML/OSBDM (UF32PBE)",       // 10
   "TBLCF - Extended TBLCF (JMxxCLD)",             // 11
   "Combined USBDM/TBLCF",                         // 12
   "Reserved #3",                                  // 13
   "Reserved #4",                                  // 14
   "Custom",                                       // 15
   };

// Version information read from BDM
static U8 currentBDMSoftwareVersion; //!< BDM Firmware version
static U8 currentBDMHardwareVersion; //!< Hardware version reported by BDM firmware
static U8 currentICPSoftwareVersion; //!< ICP Firmware version
static U8 currentICPHardwareVersion; //!< Hardware version reported by ICP firmware

// Version information available in driver DLL
static U8   availableSoftwareVersion; //!< Firware version contained in this DLL
static U8   availableHardwareVersion; //!< BMD hardware version that Firmware is intended for

//! Text description of BDM hardware detected
static const char *hardwareDescription = "Unknown";

//! \brief Windows Callback to handle dialogue events
//!
//! @param hDlg     Window Handle for the dialogue
//! @param message  The usual windows stuff
//! @param wParam
//! @param lParam
//!
static LRESULT CALLBACK Configure(HWND hDlg, UINT message,
                                  WPARAM wParam, LPARAM lParam)
{
   switch (message) {
      case WM_INITDIALOG: {
         char buff[100];
         // Text description of Hardware
         SetWindowText( GetDlgItem(hDlg, IDC_HARDWARE_TYPE),
                        hardwareDescription);
         // Create description of current firmware version in BDM
         if (currentBDMSoftwareVersion == 0xFF) // Indicates Firmware unknown (in ICP mode)
            sprintf(buff,"%s-?.? (Invalid)",
                    BDMCPUTypes[(currentBDMHardwareVersion>>6)&0x03]);
         else
            sprintf(buff,"%s-%d.%d",
                    BDMCPUTypes[(currentBDMHardwareVersion>>6)&0x03],
                    (currentBDMSoftwareVersion>>4)&0xF,currentBDMSoftwareVersion&0xF);
         SetWindowText( GetDlgItem(hDlg, IDC_CURRENT_VER), buff);

         // Description of firmware version available in driver
         sprintf(buff,"%s-%d.%d",
                 BDMCPUTypes[(currentBDMHardwareVersion>>6)&0x03],
                 (availableSoftwareVersion>>4)&0xF,availableSoftwareVersion&0xF);
         SetWindowText( GetDlgItem(hDlg, IDC_NEW_VER), buff);
         return TRUE;
      }

      case WM_COMMAND:
         switch (LOWORD(wParam)) {
            case IDOK :
               EndDialog(hDlg, LOWORD(wParam));
               return TRUE;

            case IDCANCEL :
               EndDialog(hDlg, LOWORD(wParam));
               return FALSE;

            default :
               return FALSE;
         }
   }
   return FALSE;
}

//! \brief Update BDM Firmware
//!
//! The BDM interface is reopened afterwards
//!
//! @param flashImage - The image to program into the BDM flash
//!
static int updateFirmware(FlashImage *flashImage) {
   U32u version;
   int rc;
   int numDevices;
   int retry;
   U8  icpSoftwareVersion;
   U8  icpHardwareVersion;

   print("upDateFlash: Doing ICP reboot\r\n");
   // Cause BDM to reboot into ICP mode (ignored if already in ICP mode)
   prefix(setBoot)();
   // Connection is lost so close it.
   bdm_usb_close();        // Connection is lost on reboot so close it.

   retry = 10;
   do { // Retry connection to allow for reboot, re-enumeration time
      Sleep(1000);
      numDevices=prefix(init)();
   } while ((numDevices <= 0) && (retry-- >0));

   rc = prefix(open)(0);
   if (rc != 0) {
      print("upDateFlash: Failed to re-open after reboot\r\n");
      return 1;
      }

   // Check version & in ICP mode
   version.longword    = prefix(get_version)();
   icpSoftwareVersion  = version.bytes[0]; // Software version from ICP code
   icpHardwareVersion  = version.bytes[1]; // Hardware version from ICP code

   print("upDateFlash: Checking in ICP mode\r\n");
   if ((icpSoftwareVersion != 0xFF) ||                     // Not in ICP mode or
       (icpHardwareVersion != availableHardwareVersion)) { // HW version incorrect
      print("upDateFlash: Failed to reboot to ICP mode or ICP HW incorrect\r\n");
      return 2;
   }

   // Erase Flash image area (ICP area is protected)
   print("upDateFlash: Erasing\r\n");
   rc = prefix(icp_erase)( flashImage->loadAddr,   flashImage->size );
   if (rc != 0) {
      print("upDateFlash: Failed to erase\r\n");
      return 3;
      }
   // Program the new code
   print("upDateFlash: Programming\r\n");
   rc = prefix(icp_program)( flashImage->loadAddr, flashImage->size, flashImage->data );
   if (rc != 0) {
      print("upDateFlash: Failed to program\r\n");
      return 4;
      }
   // Verify the new code
   print("upDateFlash: Verifying\r\n");
   rc = prefix(icp_verify)(  flashImage->loadAddr, flashImage->size, flashImage->data );
   if (rc != 0) {
      print("upDateFlash: Failed to verify\r\n");
      return 5;
      }

   print("upDateFlash: Doing Normal reboot (ignore errors)\r\n");
   prefix(icp_reboot)();    // Cause BDM to reboot into normal mode
   bdm_usb_close();        // Connection is lost on reboot so close it.
   retry = 10;
   do { // Retry connection to allow for reboot, re-enumeration time
      Sleep(1000);
      numDevices=prefix(init)();
   } while ((numDevices <= 0) && (retry-- >0));

   if ((numDevices <= 0) || (prefix(open)(0) != 0)) {
      print("upDateFlash: Failed to re-open after 2nd reboot\r\n");
      return 6;
      }
   print("upDateFlash: Successful\r\n");
   return 0;
}

//! \brief Display Firmware update Dialogue & optionally updates BDM Firmware
//!
//! The BDM interface is reopened
//!
//! @param hDLLInst     DLL Handle (to obtain resources)
//!
void checkFirmware(HINSTANCE hDLLInst){
int rc;
static int recurse = 0;
U32u version;
FlashImage *flashImage;

   print("checkFirmware: hDLLInst = %d\r\n", hDLLInst);

   if (hDLLInst == NULL) {
      print("checkFirmware: Error: hDLLInst handle NULL!\r\n");
      return;
   }

   // Check if already updating Flash
   if (recurse) {
      print("checkFirmware: recursing\r\n", hDLLInst);
      return;
      }
   recurse = 1;

   version.longword           = prefix(get_version)();
   currentBDMSoftwareVersion  = version.bytes[0]; // Software version of BDM code
   currentBDMHardwareVersion  = version.bytes[1]; // Hardware version from BDM code
   currentICPSoftwareVersion  = version.bytes[2]; // Software version of ICP code
   currentICPHardwareVersion  = version.bytes[3]; // Hardware version from ICP code

   print("checkFirmware: version = %4.4X\r\n", version);

   if ((currentBDMHardwareVersion & 0xC0) == 0) // ICP not available via driver unless JMxx, UFxx
      return;

   print("checkFirmware: Current BDM S/W version = %1X.%1X, Hardware version (from BDM) = %1X.%X\r\n"
         "               Current ICP S/W version = %1X.%1X, Hardware version (from ICP) = %1X.%X\r\n",
         (currentBDMSoftwareVersion>>4)&0x0F,currentBDMSoftwareVersion&0x0F,
         (currentBDMHardwareVersion>>6)&0x0F,currentBDMHardwareVersion&0x3F,
         (currentICPSoftwareVersion>>4)&0x0F,currentICPSoftwareVersion&0x0F,
         (currentICPHardwareVersion>>6)&0x0F,currentICPHardwareVersion&0x3F
         );

   if ((currentICPHardwareVersion != currentBDMHardwareVersion) &&
       (currentBDMSoftwareVersion != 0xFF)){
      print("Warning! - conflicting hardware versions reported by ICP and BDM firmware\r\n"
            " Using ICP Hardware version\r\n");
      currentBDMHardwareVersion = currentICPHardwareVersion;
   }

   // Description of hardware
   if ((currentBDMHardwareVersion & 0x3F) >
        sizeof(hardwareDescriptions)/sizeof(hardwareDescriptions[0])) {
      hardwareDescription = "Unknown";
   }
   else
      hardwareDescription = hardwareDescriptions[currentBDMHardwareVersion & 0x3F];

   print("Hardware description: %s\r\n", hardwareDescription );

   // Check each available flash image
   for (flashImage=flashImages;
        flashImage<flashImages+sizeof(flashImages)/sizeof(flashImages[0]);
        flashImage++) {
      availableSoftwareVersion = flashImage->softwareVersion;
      availableHardwareVersion = flashImage->hardwareVersion;

      print("checkFirmware: New Flash version   = %1X.%1X, Hardware version = %1X.%X\r\n",
            (availableSoftwareVersion>>4)&0x0F,availableSoftwareVersion&0x0F,
            (availableHardwareVersion>>6)&0x0F,availableHardwareVersion&0x3F);

      if ((currentBDMHardwareVersion==availableHardwareVersion) && // Correct H/W version and
          ((currentBDMSoftwareVersion<availableSoftwareVersion) || // later S/W version or
           (currentBDMSoftwareVersion == 0xFF)) ) {                // already in ICP mode
         break;
      }
   }
   if (flashImage>=flashImages+sizeof(flashImages)/sizeof(flashImages[0])) {
      // No update found
      print("checkFirmware: No update available \r\n");
      return;
   }
   rc = DialogBox(hDLLInst, (LPCTSTR)IDD_FLASH_UPDATE, 0, (DLGPROC)Configure);

   if (rc <0) { // Dialogue failure
      LPVOID lpMsgBuf;
      DWORD dw = GetLastError();

      print("checkFirmware: rc = %d\r\n", rc);

      FormatMessage(
          FORMAT_MESSAGE_ALLOCATE_BUFFER|FORMAT_MESSAGE_FROM_SYSTEM|FORMAT_MESSAGE_IGNORE_INSERTS,
          NULL,
          dw,
          MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
          (LPTSTR) &lpMsgBuf,
          0, NULL );
      print("checkFirmware: Dialogue Error %s", lpMsgBuf);
   }
   else if (rc == TRUE) {
      rc = updateFirmware(flashImage);
      if (rc != 0) {
         // Do something to alert user!
      }
   }

   recurse = 0;
}



