/*! \file
    \brief Command Line utility to test OSBDM

    \verbatim
    Copyright (C) 2008  Peter O'Donoghue

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Change History
   +====================================================================
   |  20 Oct 2008 |
   +====================================================================
    \endverbatim
*/
#include <windows.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "libusb.h"
#include "commands.h"
#include "interface.h"
#include "Common.h"
#include "CmdTable.h"

//! Type of Target
U8 targetType = T_CFVx;

//! Pause waiting for key press
void pause(void){
   printf("Press enter to continue");
   getchar();
}

//! Report Target status
static void reportState(void) {
   bdmcf_status_t BDMStatus;

   if (prefix(bdm_sts)(&BDMStatus) != 0)
      printf("Error: bdm_sts() failed\n");
   else
      printf("Status = %s, RSTO*=%s\n",
             BDMStatus.reset_detection?"Reset detected":"No Reset detected",
             BDMStatus.reset_state?"H":"L" );
}

//! Delimiters for parsing input
const char delimiters[] = "\n\t ,";

//! Reset Target to Special Mode
int resetSpecialCommand(void) {
// reset
   if (prefix(target_reset)(0) != 0)
      printf(":reset S: Failed\n");
   else
      printf(":reset S\n");
   return 0;
}

//! Reset Target to Normal Mode
int resetNormalCommand(void) {
// reset
   if (prefix(target_reset)(1) != 0)
      printf(":reset N: Failed\n");
   else
      printf(":reset N\n");
   return 0;
}

//! Connect to Target
int connectCommand(void) {
// connect
   if (prefix(resynchronize)() != 0)
      printf(":connect - Failed\n");
   else
      printf(":connect\n");
   reportState();

   return 0;
}

//! Halt Target
int haltCommand(void) {
// halt
   if (prefix(target_halt)() != 0)
      printf(":halt - Failed\n");
   else
      printf(":halt\n");
   return 0;
}

//! Start Target execution
int goCommand(void) {
// go
   if (prefix(target_go)() != 0)
      printf(":go - Failed\n");
   else
      printf(":go\n");
   return 0;
}

//! Step Target
int stepCommand(void) {
// go
   if (prefix(target_step)() != 0)
      printf(":step - Failed\n");
   else
      printf(":step\n");
   return 0;
}

//! Write a byte to target Memory
int wbCommand(void) {
// wb <addr> <data>
   char *currentToken;
   int address;
   int data;
   int count;
   U8  buff[50];

   // Write address
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%X",&address) != 1))
      return 0;

   // Data value
   for (count = 0; count < 50; count++) {
      currentToken = strtok(NULL, delimiters);
      if ((currentToken==NULL) || (sscanf(currentToken,"%X",&data) != 1))
         break;
      buff[count] = (U8) data;
   }
   if (count <= 0)
      return 0;

   prefix(write_block8)(address, count, buff);

   printf(":wb %4.4X %2.2X ...\n", address, data);
   return 0;
}

//! Write a word to target Memory
int wwCommand(void) {
// ww <addr> <data>
   char *currentToken;
   int address;
   int data;

   // Write address
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%X",&address) != 1))
      return 0;

   // Data value
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%X",&data) != 1))
      return 0;

   prefix(write_mem16)(address, data);

   printf(":ww %4.4X %4.4X\n", address, data);
   return 0;
}

//! Write a long word to target Memory
int wlCommand(void) {
// wl <addr> <data>
   char *currentToken;
   unsigned long int address;
   unsigned long int data;

   // Write address
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%lX",&address) != 1))
      return 0;

   // Data value
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%lX",&data) != 1))
      return 0;

   prefix(write_mem32)(address, data);

   printf(":wl %8.8lX %8.8lX\n", address, data);
   return 0;
}

//! Write to Target PC (HC12, HCS08 & RS08)
int wpcCommand(void) {
// wpc<addr> <data>
   char *currentToken;
   int data;

   // Data value
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%X",&data) != 1))
      return 0;


   prefix(write_creg)(0x80f, data);

   printf(":wpc %s=%X\n", getCFV2ControlRegName(0x80F), data);
   return 0;
}

//! Write to Target Register (Coldfire)
int wRegCommand(void) {
// wpc<addr> <data>
   char *currentToken;
   unsigned long int data;
   int regNo;

   // Register #
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%d",&regNo) != 1))
      return 0;
   regNo &= 0x1F;

   // Data value
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%lX",&data) != 1))
      return 0;

   prefix(write_reg)(regNo, data);

   printf(":wReg r=%d(%s)<-0x%8.8lX\n", regNo, getCFRegName(regNo), data);
   return 0;
}

//! Write to Target Debug Register (Coldfire)
int wDregCommand(void) {
// wpc<addr> <data>
   char *currentToken;
   unsigned long int data;
   int regNo;

   // Register #
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%X",&regNo) != 1))
      return 0;
   regNo &= 0x1F;

   // Data value
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%lX",&data) != 1))
      return 0;

   prefix(write_dreg)(regNo, data);

   printf(":wDreg r=0x%X(%s)<-0x%lX\n", regNo, getCFDebugRegName(regNo), data);
   return 0;
}

//! Write to Target Control Register (Coldfire)
int wCregCommand(void) {
// wpc<addr> <data>
   char *currentToken;
   unsigned long int data;
   int regNo;
   // Register #
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%X",&regNo) != 1))
      return 0;
   // Data value
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%lX",&data) != 1))
      return 0;

   prefix(write_creg)(regNo, data);

   printf(":wCreg r=0x%X(%s)<-0x%8.8lX\n",
             regNo, getCFV2ControlRegName(regNo), data);
   return 0;
}

//! Read from Target Register (Coldfire)
int rRegCommand(void) {
// wpc<addr> <data>
   char *currentToken;
   unsigned long int data;
   int regNo;

   // Register #
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%d",&regNo) != 1))
      return 0;
   regNo &= 0x1F;

   prefix(read_reg)(regNo, &data);

   printf(":Dreg r=%d(%s)->%8.8lX\n", regNo, getCFRegName(regNo), data);
   return 0;
}

//! Read from Target Debug Register (Coldfire)
int rDregCommand(void) {
// wpc<addr> <data>
   char *currentToken;
   unsigned long int data;
   int regNo;
   // Register #
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%X",&regNo) != 1))
      return 0;
   regNo &= 0x1F;

   prefix(read_dreg)(regNo, &data);

   printf(":rDreg r=0x%X(%s)<-0x%8.8lX\n", regNo, getCFDebugRegName(regNo), data);
   return 0;
}

//! Read from Target Control Register (Coldfire)
int rCregCommand(void) {
// wpc<addr> <data>
   char *currentToken;
   unsigned long int data;
   int regNo;

   // Register #
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%X",&regNo) != 1))
      return 0;

   prefix(read_creg)(regNo, &data);

   printf(":rCreg r=0x%X(%s)<-0x%lX\n", regNo, getCFV2ControlRegName(regNo), data);

   return 0;
}

#if 0
//! Testing
int testCregCommand(void) {
// wpc<addr> <data>
   int dataH, dataL;
   int regNo;
   static const char *names[] = {"_A7","VBR","CPUCR","?","?","?","?","?",
                                "?","?","?","?","?","?","SR","PC",
   };

   for( regNo = 0; regNo<=15; regNo++) {
      prefix(write_creg)(regNo, 0xFFFFFFFF);
      dataH = prefix(read_creg)(regNo);
      prefix(write_creg(regNo, 0x00000000);
      dataL = prefix(read_creg)(regNo);

      printf(":testCreg%d(%5s) 0x00000000->%8X, 0xFFFFFFFF->%8X\n",
               regNo, names[regNo], dataL, dataH);
   }

   return 0;
}
#endif

//! Write block to Target memory
int wblockCommand(void) {
// wb <addr> <data>
   char *currentToken;
   int address;
   int size;
   int data;
   int startData;
   int sub;
   unsigned char dataBlock[1024];

   // Write start address
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%X",&address) != 1))
      return 0;

   // Number of bytes to write
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%X",&size) != 1))
      return 0;
   if (size > sizeof(dataBlock))
      size = sizeof(dataBlock);

   // Data value
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%X",&data) != 1))
      return 0;

   startData = data;
   for (sub = 0; sub <size; sub++) {
      dataBlock[sub]  = data;
      data++;
   }

   prefix(write_block8)(address, size, dataBlock);

   printf(":wblock [%4.4X-%4.4X]=%2.2X...%2.2X\n",
            address, address+size-1, startData, data-1);
   return 0;
}

//! Read byte from Target memory
int rbCommand(void) {
// rb <addr>
   char *currentToken;
   int address;
   U8 data;

   // Read address
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%X",&address) != 1))
      return 0;

   prefix(read_mem8)(address, &data);

   printf(":rb %4.4X %2.2X\n", address, data);
   return 0;
}

//! Read word from Target memory
int rwCommand(void) {
// rw <addr>
   char *currentToken;
   int  address;
   int  data;

   // Read address
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%X",&address) != 1))
      return 0;

   prefix(read_mem16)(address, &data);

   printf(":rw %4.4X %4.4X\n", address, data);
   return 0;
}

//! Read long word from Target memory
int rlCommand(void) {
// rw <addr>
   char *currentToken;
   unsigned long int  address;
   unsigned long int  data;

   // Read address
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%lX",&address) != 1))
      return 0;

   prefix(read_mem32)(address, &data);

   printf(":rw %8.8lX %8.8lX\n", address, data);
   return 0;
}

//! Read block from Target memory
int rblockCommand(void) {
// rb <addr> <data>
   char *currentToken;
   int address;
   int size;
   int sub;
   unsigned char dataBlock[1024];

   // Start address
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%X",&address) != 1))
      return 0;

   // Number of bytes to read
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%X",&size) != 1))
      return 0;

   if (size > sizeof(dataBlock))
      size = sizeof(dataBlock);

   for (sub = 0; sub <size; sub++)
      dataBlock[sub] = 0x00;


   prefix(read_block8)(address, size, dataBlock);

   printf(":rblock [%4.4X-%4.4X]\n", address, address+size-1);
   for (sub=0; sub<size; sub++){
      if ((sub&0x0F) == 0)
         printf("\n%4.4X:", address+sub);
      printf("%2.2X ", dataBlock[sub]);
   }
   printf("\n");

   return 0;
}

int testBlockCommand(void) {
// tBlock adder size repeatcount

   char *currentToken;
   int address;
   int repeatCount;
   int maxRepeat;
   int size;
   int sub;
   int errorCount = 0;
   unsigned char dataWriteBlock[20000];
   unsigned char dataReadBlock[20000];
   time_t    time_start;
   time_t    time_end;

   // Start address
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%X",&address) != 1))
      return 0;

   // Number of bytes to read
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%X",&size) != 1))
      return 0;

   // Number of times to repeat
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%d",&maxRepeat) != 1))
      maxRepeat = 1;

   if (size > sizeof(dataWriteBlock))
      size = sizeof(dataWriteBlock);

   printf(":tblock [%4X-%4X], %d times\n", address, address+size-1, maxRepeat);

   time(&time_start);

   for (repeatCount=0;  repeatCount<maxRepeat; repeatCount++) {
      int data = rand()&0xFF;

      if ((repeatCount%20) == 0)
         printf("%6.6d:",repeatCount);
      printf("%2.2X,", data);
      if (((repeatCount%20) == 19) && (repeatCount<maxRepeat))
         printf("\n");

      for (sub = 0; sub <size; sub++) {
         dataWriteBlock[sub]  = data;
         data                += 7;
      }

      prefix(write_block8)(address, size,dataWriteBlock);

      prefix(read_block8)(address, size, dataReadBlock);

      for (sub = 0; sub <size; sub++)
         if (dataWriteBlock[sub] != dataReadBlock[sub]) {
            printf("addr = %4X, wrote %2.2X, read %2.2X\n",
                   address+sub, dataWriteBlock[sub], dataReadBlock[sub]);
            errorCount++;
         }
#if 0
      printf(":tblock [%4.4X-%4.4X]\n", address, address+size-1);
      for (sub=0; sub<size; sub++){
         if ((sub&0x0F) == 0)
            printf("\n%4.4X:", address+sub);
         printf("%2.2X ", dataReadBlock[sub]);
      }
      printf("\n");
#endif
   }

   time(&time_end);
   printf("\n");
   printf("Started:   %s",  asctime(localtime(&time_start)));
   printf("Completed: %s\n",  asctime(localtime(&time_end)));

   printf("\n:tblock [%4X-%4X], error count = %d\n",
          address, address+size-1, errorCount);

   return 0;
}

//! Read Registers from Target
int getRegsCommand(void) {
   int regNo;
   unsigned long int regVal;

   printf("gregs\n");

   for( regNo = 0; regNo<=15; regNo++) {
      prefix(read_reg)(regNo, &regVal);
      printf("%3s =%8.8lX, ", getCFRegName(regNo), regVal);
      if ((regNo&0x03)== 0x03)
         printf("\n");
   }

   prefix(read_creg)(0x80e, &regVal);
   printf("%3s =%C-%C%C-%d---%C%C%C%C%C,         ",
          getCFV2ControlRegName(0x80e),
         (regVal&0x8000)?'T':'-',
         (regVal&0x2000)?'S':'-',
         (regVal&0x1000)?'M':'-',
         (U8)(regVal>>8)&0x7,
         (regVal&0x0010)?'X':'-',
         (regVal&0x0008)?'N':'-',
         (regVal&0x0004)?'Z':'-',
         (regVal&0x0002)?'V':'-',
         (regVal&0x0001)?'C':'-' );
   prefix(read_creg)(0x800, &regVal);
   printf(" %3s =%8.8lX, \n", getCFV2ControlRegName(0x800), (U32)regVal);
   prefix(read_creg)(0x801, &regVal);
   regVal &= 0xFFF00000;
   printf("%3s =%8.8lX, ", getCFV2ControlRegName(0x801), (U32)regVal);
   prefix(read_creg)(0x80f, &regVal);
   printf("%3s =%8.8lX\n", getCFV2ControlRegName(0x80f), (U32)regVal);
   prefix(read_creg)(0xC04, &regVal);
   printf("%3s =%8.8lX,  ", getCFV2ControlRegName(0xC04), (U32)regVal);
   prefix(read_creg)(0xC05, &regVal);
   printf("%3s =%8.8lX\n", getCFV2ControlRegName(0xC05), (U32)regVal);

   return 0;
}

//! Delay a while
int delayCommand(void) {
// delay <1/10 seconds>
   char *currentToken;
   int delay;

   // Read delay
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%d",&delay) != 1))
      return 0;

   Sleep(100*delay);
   printf(":delay %d\n", delay);
   return 0;
}

//! Set target Type
int setTargetCommand(void) {
// setTarget <value>
   char *currentToken;
   int   value;

   // Read control value to set
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%d",&value) != 1))
      return 1;

   targetType = (U8) value;

   if (prefix(set_target_type)((U8)value) != 0)
      printf(":setTarget %d - Failed\n", value);
   else
      printf(":setTarget %d\n", value);
   return 0;
}

//! Set Target to re-boot into ICP mode
int setBootCommand(void) {
// setBoot
   prefix(setBoot)();

   printf("setBootCommand Complete\n");
   return 0;
}

//! Send debug command to BDM
int debugCommand(void) {
// debug <control_value>
   char *currentToken;
   int value;
   unsigned char usb_data[20] = {0};

   // Read control value to set
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%d",&value) != 1))
      return 1;

   usb_data[0]=2;  /* receive 2 bytes */
   usb_data[1]=CMD_DEBUG;
   usb_data[2]=value;

   bdm_debug(usb_data);

   printf(":debug %2.2X => %2.2X(%2.2d), %2.2X(%2.2d), "
          "%2.2X(%2.2d), %2.2X(%2.2d) \n", value,
          usb_data[0], usb_data[0], usb_data[1], usb_data[1],
          usb_data[2], usb_data[2], usb_data[3], usb_data[3] );
   reportState();

   return 0;
}

//! Exit program
int exitCommand(void) {
   return 1;
}

//! Do nothing
int commentCommand( void ) {
   return 0;
}

//! Usage message
static const char usageText[] =
   "connect                               - connect to target\n\r"
   "debug <value>                         - debug commands\n\r"
   "delay <time>                          - pause a while\n\r"
   "db <addr>                             - read byte\n\r"
   "dw <addr>                             - read word\n\r"
   "dl <addr>                             - read longword (CFV1 only)\n\r"
   "exit                                  - exit program\n\r"
   "go                                    - start from current PC\n\r"
   "gs                                    - read status register\n\r"
   "halt                                  - halt the target\n\r"
   "help                                  - display help\n\r"
   "regs                                  - print out registers\n\r"
   "reset                                 - Reset to bdm mode\n\r"
   "resetn                                - Reset to normal mode\n\r"
   "rb <addr>                             - read byte\n\r"
   "rblock <addr> <size>                  - read block\n\r"
   "rw <addr>                             - read word\n\r"
   "rl <addr>                             - read longword (CFV1 only)\n\r"
   "setboot                               - set USBDM module to ICP mode\n\r"
   "settarget <target>                    - 0=HCS12, 1=HCS08, 2=RS08, 3=CFV1\n\r"
   "speed <Hz>                            - set speed \n\r"
   "step                                  - execute a single instruction\n\r"
   "tblock <addr> <size> <repeatcount>    - random RAM write/read block test\n\r"
   "trim <params>                         - calculate clock trim value\n\r"
   "wait <time>                           - pause a while\n\r"
   "wb <addr> <data>                      - write byte\n\r"
   "wblock <addr> <size> <data>           - write block\n\r"
   "wc <value>                            - write control\n\r"
   "wpc <value>                           - write to PC\n\r"
   "ww <addr><value>                      - write word\n\r"
   "wl <addr><value>                      - write longword (CFV1 only)\n\r"
   ;

//! Help command
int helpCommand(void) {
   fprintf(stderr,"%s",usageText);
   return 0;
}

//! Command Entry structure
typedef struct {
   char *command;       //!< Command name (string)
   int (*func)(void);   //!< Command function to execute
} CommandEntry;

//! Command Table
CommandEntry commandTable[] = {
   {"connect",   connectCommand},
   {"debug",     debugCommand},
   {"delay",     delayCommand},
   {"db",        rbCommand},
   {"dw",        rwCommand},
   {"dl",        rlCommand},
   {"exit",      exitCommand},
   {"go",        goCommand},
   {"halt",      haltCommand},
   {"help",      helpCommand},
   {"regs",      getRegsCommand},
   {"reset",     resetSpecialCommand},
   {"resetn",    resetNormalCommand},
   {"rb",        rbCommand},
   {"rblock",    rblockCommand},
   {"rw",        rwCommand},
   {"rl",        rlCommand},
   {"setboot",   setBootCommand},
   {"settarget", setTargetCommand},
   {"step",      stepCommand},
   {"tblock",    testBlockCommand},
   {"wait",      delayCommand},
   {"wb",        wbCommand},
   {"wblock",    wblockCommand},
   {"wpc",       wpcCommand},
   {"ww",        wwCommand},
   {"wl",        wlCommand},
   {"?",         helpCommand},
   {"wreg",      wRegCommand},
   {"wcreg",     wCregCommand},
   {"wdreg",     wDregCommand},
   {"rreg",      rRegCommand},
   {"rcreg",     rCregCommand},
   {"rdreg",     rDregCommand},
};

//! Search for & execute command
//!
int execute(char *commandLine){
   char           *currentToken;
   CommandEntry   *commandEntry;

   // Remove comments
   currentToken = strstr(commandLine, "//");
   if (currentToken != NULL){
      *currentToken = '\0';
   }
   currentToken = strtok(commandLine, delimiters);

   if (currentToken == NULL)
      return 0;

   for( commandEntry = commandTable;
        commandEntry < commandTable+sizeof(commandTable)/sizeof(commandTable[0]);
        commandEntry++ ){
      if (strcmp(currentToken,commandEntry->command)==0){
         return commandEntry->func();
      }
   }
   printf("Unknown command \r\n");
   return 0;
}

//! Mainline
int main(void){
int numDevices;
char commandBuffer[100];
int errCode;

   HINSTANCE hInst = GetModuleHandle(0);

   dll_initialize(hInst);

   numDevices = prefix(init)();

   if (numDevices == 0) {
      fprintf(stderr, "No devices Found\n");
      return 1;
   }
   prefix(open)(0);

   reportState();

   while (gets(commandBuffer) != NULL){
      printf("Executing:%s\n", commandBuffer);
      errCode = execute(commandBuffer);
      if ((errCode != 0)) {
         printf("Failed\n");
         break;
      }
   }
   prefix(close)();
   dll_uninitialize();
   return 0;
}

#if 0
int main2(void) {
   char commandBuffer[100];
   FILE *commandFile;
   int errCode;

   commandFile = fopen("MC9S08QG8.trim", "rt");
   if (commandFile == NULL){
      fprintf(stderr,"Unable to open trim file");
      return 0;
   }
   dev = prefix(Connect)();
   if (dev == NULL) {
      printf("error: device not found!\n");
      return 0;
   }

   while (gets(commandBuffer/*, 80, commandFile*/) != NULL){
      printf("Executing:%s\n", commandBuffer);
      errCode = execute(commandBuffer);
      if ((errCode != 0)) {
         printf("Failed\n");
         break;
      }
   }

   //========================================================
   //CLEANUP:
   if (dev != NULL) {
      usb_release_interface(dev, 0);
      usb_close(dev);
   }

   if (commandFile != NULL)
      fclose(commandFile);

  return 0;
}

#endif
