/*! \file
    \brief Command Line utility to test OSBDM

    \verbatim
    Copyright (C) 2008  Peter O'Donoghue

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Change History
   +====================================================================
   |  20 Oct 2008 |
   +====================================================================
    \endverbatim
*/
#include <windows.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "libusb.h"
#include "commands.h"
#include "interface.h"
#include "Common.h"

//! Type of Target
U8 targetType = T_OFF;

//! Pause waiting for key press
void pause(void){
   printf("Press enter to continue");
   getchar();
}

//! Report Target status
static void reportState(void) {
   bdm_status_t BDMStatus;
   float speed;
   static const char *connectionStates[] = {"Not Connected", "Sync Speed", "Guess Speed", "Manual Speed"};

   if (prefix(bdm_sts)(&BDMStatus) != 0)
      printf("Error: bdm_sts() failed\n");
   else
      printf("Status = %s, %s, %s\n", BDMStatus.ackn_state?"Ackn":"Wait",
                                      connectionStates[BDMStatus.connection_state],
                                      BDMStatus.reset_state?"Reset":"No Reset" );
   speed = prefix(get_speed)();
   printf("Speed = %.2f MHz (%.0f ticks, sync=%.1f us)\n",
          speed, (60.0 * 2 * 128)/speed, (2 * 128)/speed);
}

//! Delimiters for parsing input
const char delimiters[] = "\n\t ,";

//! Reset Target to Special Mode
int resetSpecialCommand(void) {
// reset
   if (prefix(target_reset)(0) != 0)
      printf(":reset S: Failed\n");
   else
      printf(":reset S\n");
   return 0;
}

//! Reset Target to Normal Mode
int resetNormalCommand(void) {
// reset
   if (prefix(target_reset)(1) != 0)
      printf(":reset N: Failed\n");
   else
      printf(":reset N\n");
   return 0;
}

//! Connect to Target
int connectCommand(void) {
// connect
   if (prefix(target_sync)() != 0)
      printf(":connect - Failed\n");
   else
      printf(":connect\n");
   reportState();

   return 0;
}

//! Halt Target
int haltCommand(void) {
// halt
   if (prefix(target_halt)() != 0)
      printf(":halt - Failed\n");
   else
      printf(":halt\n");
   return 0;
}

//! Start Target execution
int goCommand(void) {
// go
   if (prefix(target_go)() != 0)
      printf(":go - Failed\n");
   else
      printf(":go\n");
   return 0;
}

//! Step Target
int stepCommand(void) {
// go
   if (prefix(target_step)() != 0)
      printf(":step - Failed\n");
   else
      printf(":step\n");
   return 0;
}

//! Write a byte to target Memory
int wbCommand(void) {
// wb <addr> <data>
   char *currentToken;
   int address;
   int data;
   int count;
   U8  buff[50];

   // Write address
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%X",&address) != 1))
      return 0;

   // Data value
   for (count = 0; count < 50; count++) {
      currentToken = strtok(NULL, delimiters);
      if ((currentToken==NULL) || (sscanf(currentToken,"%X",&data) != 1))
         break;
      buff[count] = (U8) data;
   }
   if (count <= 0)
      return 0;

   if (targetType == T_CFV1) {
      prefix(write_mem)(address, count, 1, buff);
      }
   else {
      int sub;
      for (sub = 0; sub < count; sub++)
         prefix(write_byte)(address, buff[sub]);
   }
   printf(":wb %4.4X %2.2X ...\n", address, data);
   return 0;
}

//! Write a word to target Memory
int wwCommand(void) {
// ww <addr> <data>
   char *currentToken;
   int address;
   int data;
   U8 buff[2];

   // Write address
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%X",&address) != 1))
      return 0;

   // Data value
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%X",&data) != 1))
      return 0;

   if (targetType == T_CFV1) {
      buff[0] = (U8)(data>>8); // Little Endian in target memory!
      buff[1] = (U8)(data);
      prefix(write_mem)(address, 2, 2, buff);
      }
   else
      prefix(write_word)(address, data);

   printf(":ww %4.4X %4.4X\n", address, data);
   return 0;
}

//! Write a long word to target Memory
int wlCommand(void) {
// wl <addr> <data>
   char *currentToken;
   int address;
   int data;
   U8 buff[4];

   // Write address
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%X",&address) != 1))
      return 0;

   // Data value
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%X",&data) != 1))
      return 0;

   if (targetType == T_CFV1) {
      buff[0] = (U8)(data>>24); // Big Endian in target memory!
      buff[1] = (U8)(data>>16);
      buff[2] = (U8)(data>>8);
      buff[3] = (U8)(data);
      prefix(write_mem)(address, 4, 4, buff);
      }
   else {
      printf("Only supported on CFV1 Targets\n");
      return 0;
   }


   printf(":ww %4.4X %8.8X\n", address, data);
   return 0;
}

//! Write to Target PC (HC12, HCS08 & RS08)
int wpcCommand(void) {
// wpc<addr> <data>
   char *currentToken;
   int data;

   // Data value
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%X",&data) != 1))
      return 0;

   if (targetType == T_CFV1)
      prefix(write_creg)(15, data);
   else
      prefix(write_reg_pc)(data);
   printf(":wpc %X\n", data);
   return 0;
}

//! Write to Target Register (Coldfire)
int wRegCommand(void) {
// wpc<addr> <data>
   char *currentToken;
   int data;
   int regNo;
   static const char *names[] = {"D0","D1","D2","D3","D4","D5","D6","D7",
                                 "A0","A1","A2","A3","A4","A5","A6","A7",
                                 "?","?","?","?","?","?","?","?",
                                 "?","?","?","?","?","?","?","?",
   };

   // Register #
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%d",&regNo) != 1))
      return 0;
   regNo &= 0x1F;

   // Data value
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%X",&data) != 1))
      return 0;

   prefix(write_reg)(regNo, data);

   printf(":wReg r=%d(%s)<-%X\n", regNo, names[regNo], data);
   return 0;
}

//! Write to Target Debug Register (Coldfire)
int wDregCommand(void) {
// wpc<addr> <data>
   char *currentToken;
   int data;
   int regNo;
   static const char *names[] = {"CSR","XCSR","CSR2","CSR3","?","BAAR","AATR","TDR",
                                "PBR0","PBMR","?","?","ABHR","ABLR","DBR","DBMR",
                                "?","?","?","?","?","?","?","?",
                                "PBR1","PBR2","PBR3","?","?","?","?","?",
   };

   // Register #
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%d",&regNo) != 1))
      return 0;
   regNo &= 0x1F;

   // Data value
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%X",&data) != 1))
      return 0;

   prefix(write_dreg)(regNo, data);

   printf(":wDreg r=%d(%s)<-%X\n", regNo, names[regNo], data);
   return 0;
}

//! Write to Target Control Register (Coldfire)
int wCregCommand(void) {
// wpc<addr> <data>
   char *currentToken;
   int data;
   int regNo;
   static const char *names[] = {"_A7","VBR","CPUCR","?","?","?","?","?",
                                 "?","?","?","?","?","?","SR","PC",
                                 "?","?","?","?","?","?","?","?",
                                 "?","?","?","?","?","?","?","?",
   };

   // Register #
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%d",&regNo) != 1))
      return 0;
   regNo &=0x1F;
   // Data value
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%X",&data) != 1))
      return 0;

   prefix(write_creg)(regNo, data);

   printf(":wCreg r=%d(%s)<-%X\n", regNo, names[regNo], data);
   return 0;
}

//! Read from Target Register (Coldfire)
int rRegCommand(void) {
// wpc<addr> <data>
   char *currentToken;
   int data;
   int regNo;

   static const char *names[] = {"D0","D1","D2","D3","D4","D5","D6","D7",
                                 "A0","A1","A2","A3","A4","A5","A6","A7",
                                 "?","?","?","?","?","?","?","?",
                                 "?","?","?","?","?","?","?","?",
   };

   // Register #
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%d",&regNo) != 1))
      return 0;
   regNo &= 0x1F;

   data = prefix(read_reg)(regNo);


   printf(":Dreg r=%d(%s)->%X\n", regNo, names[regNo], data);
   return 0;
}

//! Read from Target Debug Register (Coldfire)
int rDregCommand(void) {
// wpc<addr> <data>
   char *currentToken;
   int data;
   int regNo;
   static const char *names[] = {"CSR","XCSR","CSR2","CSR3","?","BAAR","AATR","TDR",
                                "PBR0","PBMR","?","?","ABHR","ABLR","DBR","DBMR",
                                "?","?","?","?","?","?","?","?",
                                "PBR1","PBR2","PBR3","","","","","",
   };

   // Register #
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%d",&regNo) != 1))
      return 0;
   regNo &= 0x1F;

   data = prefix(read_dreg)(regNo);

   printf(":rDreg r=%d(%s)->%X\n", regNo, names[regNo], data);
   return 0;
}

//! Read from Target Control Register (Coldfire)
int rCregCommand(void) {
// wpc<addr> <data>
   char *currentToken;
   int data;
   int regNo;
   static const char *names[] = {"_A7","VBR","CPUCR","?","?","?","?","?",
                                "?","?","?","?","?","?","SR","PC",
                                "?","?","?","?","?","?","?","?",
                                "?","?","?","?","?","?","?","?",
   };

   // Register #
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%d",&regNo) != 1))
      return 0;
   regNo &=0x1F;

   data = prefix(read_creg)(regNo);

   printf(":cRreg r=%d(%s)->%X\n", regNo, names[regNo], data);
   return 0;
}

//! Testing
int testCregCommand(void) {
// wpc<addr> <data>
   int dataH, dataL;
   int regNo;
   static const char *names[] = {"_A7","VBR","CPUCR","?","?","?","?","?",
                                "?","?","?","?","?","?","SR","PC",
   };

   for( regNo = 0; regNo<=15; regNo++) {
      prefix(write_creg)(regNo, 0xFFFFFFFF);
      dataH = prefix(read_creg)(regNo);
      prefix(write_creg)(regNo, 0x00000000);
      dataL = prefix(read_creg)(regNo);

      printf(":testCreg%d(%5s) 0x00000000->%8X, 0xFFFFFFFF->%8X\n",
               regNo, names[regNo], dataL, dataH);
   }

   return 0;
}

//! Write block to Target memory
int wblockCommand(void) {
// wb <addr> <data>
   char *currentToken;
   int address;
   int size;
   int data;
   int sub;
   unsigned char dataBlock[1024];

   // Write start address
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%X",&address) != 1))
      return 0;

   // Number of bytes to write
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%X",&size) != 1))
      return 0;
   if (size > sizeof(dataBlock))
      size = sizeof(dataBlock);

   // Data value
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%X",&data) != 1))
      return 0;

   for (sub = 0; sub <size; sub++)
      dataBlock[sub] = data++;

   if (targetType == T_CFV1) {
      prefix(write_mem)(address, size, 4, dataBlock);
      }
   else
      prefix(write_block)(address, size, dataBlock);

   printf(":wblock [%4.4X-%4.4X]=%2.2X\n", address, address+size-1, data);
   return 0;
}

//! Read block from Target memory
int rbCommand(void) {
// rb <addr>
   char *currentToken;
   int address;
   U8 data;
   char buff[1];

   // Read address
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%X",&address) != 1))
      return 0;

   if (targetType == T_CFV1) {
      prefix(read_mem)(address, 1, 1, buff);
      data = buff[0];
      }
   else
      data = prefix(read_byte)(address);

   printf(":rb %4.4X %2.2X\n", address, data);
   return 0;
}

//! Read word from Target memory
int rwCommand(void) {
// rw <addr>
   char *currentToken;
   int  address;
   int  data;
   char buff[2];

   // Read address
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%X",&address) != 1))
      return 0;

   if (targetType == T_CFV1) {
      prefix(read_mem)(address, 2, 2, buff);
      data = *(U16 *)buff;
      }
   else
      data = prefix(read_word)(address);

   printf(":rw %4.4X %4.4X\n", address, data);
   return 0;
}

//! Read long word from Target memory
int rlCommand(void) {
// rw <addr>
   char *currentToken;
   int  address;
   int  data;
   char buff[4];

   // Read address
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%X",&address) != 1))
      return 0;

   if (targetType == T_CFV1) {
      prefix(read_mem)(address, 4, 4, buff);
      data = *(U32 *)buff;
      }
   else {
      printf("Not supported except on CFV1\n");
      return 0;
   }

   printf(":rw %4.4X %8.8X\n", address, data);
   return 0;
}

//! Read block from Target memory
int rblockCommand(void) {
// rb <addr> <data>
   char *currentToken;
   int address;
   int size;
   int sub;
   unsigned char dataBlock[1024];

   // Start address
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%X",&address) != 1))
      return 0;

   // Number of bytes to read
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%X",&size) != 1))
      return 0;

   if (size > sizeof(dataBlock))
      size = sizeof(dataBlock);

   for (sub = 0; sub <size; sub++)
      dataBlock[sub] = 0x00;

   if (targetType == T_CFV1) {
      prefix(read_mem)(address, size, 4, dataBlock);
      }
   else
      prefix(read_block)(address, size, dataBlock);

   printf(":rblock [%4.4X-%4.4X]\n", address, address+size-1);
   for (sub=0; sub<size; sub++){
      if ((sub&0x0F) == 0)
         printf("\n%4.4X:", address+sub);
      printf("%2.2X ", dataBlock[sub]);
   }
   printf("\n");

   return 0;
}

int testBlockCommand(void) {
// tBlock adder size repeatcount

   char *currentToken;
   int address;
   int repeatCount;
   int maxRepeat;
   int size;
   int sub;
   int errorCount = 0;
   unsigned char dataWriteBlock[20000];
   unsigned char dataReadBlock[20000];
   time_t    time_start;
   time_t    time_end;

   // Start address
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%X",&address) != 1))
      return 0;

   // Read end address
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%X",&size) != 1))
      return 0;

   if (size < address)
      return 0;

   size -= address;
   size++;

   // Number of times to repeat
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%d",&maxRepeat) != 1))
      maxRepeat = 1;

   if (size > sizeof(dataWriteBlock))
      size = sizeof(dataWriteBlock);

   printf(":tblock [0x%4.4X-0x%4.4X] (0x%X bytes), %d times\n", address, address+size-1, size, maxRepeat);

   if (((address & 0x03)!= 0) || ((size & 0x03)!= 0)) {
      printf(":tblock - Alignment error\n");
      return 0;
   }

   time(&time_start);

   for (repeatCount=0;  repeatCount<maxRepeat; repeatCount++) {
      int data = rand()&0xFF;

      if ((repeatCount%20) == 0)
         printf("%6.6d:",repeatCount);
      printf("%2.2X,", data);
      if (((repeatCount%20) == 19) && (repeatCount<maxRepeat))
         printf("\n");

      for (sub = 0; sub <size; sub++)
         dataWriteBlock[sub] = data++;
      if (targetType == T_CFV1) {
         prefix(write_mem)(address, size, 4, dataWriteBlock);
         }
      else
         prefix(write_block)(address, size, dataWriteBlock);
      if (targetType == T_CFV1)
         prefix(read_mem)(address, size, 4, dataReadBlock);
      else
         prefix(read_block)(address, size, dataReadBlock);

      for (sub = 0; sub <size; sub++)
         if (dataWriteBlock[sub] != dataReadBlock[sub]) {
            printf("addr = %4X, wrote %2.2X, read %2.2X\n",
                   address+sub, dataWriteBlock[sub], dataReadBlock[sub]);
            errorCount++;
         }
#if 0
      printf(":tblock [%4.4X-%4.4X]\n", address, address+size-1);
      for (sub=0; sub<size; sub++){
         if ((sub&0x0F) == 0)
            printf("\n%4.4X:", address+sub);
         printf("%2.2X ", dataReadBlock[sub]);
      }
      printf("\n");
#endif
   }

   time(&time_end);
   printf("\n");
   printf("Started:   %s",  asctime(localtime(&time_start)));
   printf("Completed: %s\n",  asctime(localtime(&time_end)));

   printf("\n:tblock [%4X-%4X], error count = %d\n",
          address, address+size-1, errorCount);

   return 0;
}

//! Read Status Value from Target
int getStatusCommand(void) {
   U8 targetStatus = 0x00;

   targetStatus = prefix(read_status)();               // Read BDMSTS register

   printf("BDMSCR = %2X\n", targetStatus);
   reportState();

   return 0;
}

//! Read Registers from Target
int getRegsCommand(void) {

   registers_t regs;


   printf("gregs\n");

   switch (targetType) {
      case T_HC12 :
         prefix(read_regs)(&regs);
         printf("PC=%4x, SP=%4X, X=%4X, Y=%4X, D=%4X, CCR=%2X\n",
                regs.hc12.pc, regs.hc12.sp, regs.hc12.ix,
                regs.hc12.iy, regs.hc12.d,  regs.hc12.ccr);
         break;
      case T_HCS08 :
         prefix(read_regs)(&regs);
         printf("PC=%4x, SP=%4X, H:X=%4X, A=%2X, CCR=%2X\n",
                regs.hcs08.pc, regs.hcs08.sp, regs.hcs08.hx,
                regs.hcs08.a,  regs.hcs08.ccr);
         break;
      case T_RS08 :
         // ToDo
         prefix(read_regs)(&regs);
         printf("PC=%4x, SP=%4X, H:X=%4X, A=%2X, CCR=%2X\n",
                regs.hcs08.pc, regs.hcs08.sp, regs.hcs08.hx,
                regs.hcs08.a,  regs.hcs08.ccr);
         break;
      case T_CFV1 : {
         int regNo;
         U32 regVal;
         for( regNo = 0; regNo<=7; regNo++) {
            regVal = prefix(read_reg)(regNo);
            printf("D%1d =%8.8lX, ", regNo, regVal);
            if ((regNo&0x03)== 0x03)
               printf("\n");
         }
         for( regNo = 8; regNo<=15; regNo++) {
            regVal = prefix(read_reg)(regNo);
            printf("A%1d =%8.8lX, ", regNo-8, regVal);
            if ((regNo&0x03)== 0x03)
               printf("\n");
         }
         regVal = prefix(read_creg)(15);
         printf("PC =%8.8X, ", (U16)regVal);
         regVal = prefix(read_creg)(14);
         printf("SR =%C-%C%C-%d---%C%C%C%C%C,         ",
                (regVal&0x8000)?'T':'-',
                (regVal&0x2000)?'S':'-',
                (regVal&0x1000)?'M':'-',
                (U8)(regVal>>8)&0x7,
                (regVal&0x0010)?'X':'-',
                (regVal&0x0008)?'N':'-',
                (regVal&0x0004)?'Z':'-',
                (regVal&0x0002)?'V':'-',
                (regVal&0x0001)?'C':'-' );
         regVal = prefix(read_creg)(0);
         printf("_SP=%8.8lX, \n", (U32)regVal);
         regVal = prefix(read_creg)(1)&0x00F00000;
         printf("VBR=%8.8lX, ", (U32)regVal);
         regVal = prefix(read_creg)(2)&0xFA000000;
         printf("CPUCR=%8.8lX\n", (U32)regVal);
      }
         break;
   }

   return 0;
}

//! Delay a while
int delayCommand(void) {
// delay <1/10 seconds>
   char *currentToken;
   int delay;

   // Read delay
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%d",&delay) != 1))
      return 0;

   Sleep(100*delay);
   printf(":delay %d\n", delay);
   return 0;
}

//! Trim Target Clock
int trimCommand(void) {
//trim <mode=8/9 bit> <ICSTRM> <ICSSC> <NV_ICSTRM> <NV_FTRIM> <Target_Frequency>
   char *currentToken;
   int mode;
   int addrICSSC, addrICSTRIM, addrNV_FTRIM, addrNV_ICSTRM;
   float targetFrequency;
   float currentSpeed;
#define SEARCH_OFFSET (10)

   // Read mode
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%d",&mode) != 1))
      return 0;

   // Read ICSTRIM address
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%X",&addrICSTRIM) != 1))
      return 0;

   // Read ICSSC address
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%X",&addrICSSC) != 1))
      return 0;

   // Read NV_ICSTRM address
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%X",&addrNV_ICSTRM) != 1))
      return 0;

   // Read NV_TRIM address
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%X",&addrNV_FTRIM) != 1))
      return 0;

   // Read target frequency
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%f",&targetFrequency) != 1))
      return 0;
   int currentTrimValue = 0x000; // 9-bit 1/2 way
   int trialBit;

   printf("trim <mode=8/9 bit> <ICSTRM> <ICSSC> <NV_ICSTRM> <NV_FTRIM> <Target_Frequency>\n");
   printf("          %s         %4.4X    %4.4X     %4.4X      %4.4X       %7.2f\n",
          mode?"8":"9",
          addrICSTRIM,
          addrICSSC,
          addrNV_ICSTRM,
          addrNV_FTRIM,
          targetFrequency
          );
   printf("1. Binary search\n");
   for (trialBit = 0x100; trialBit > 0x0; trialBit >>= 1){
      // Add in the trial bit & check speed
      currentTrimValue |= trialBit;
      prefix(write_byte)(addrICSSC,   currentTrimValue&0x01);   //  1-bit
      prefix(write_byte)(addrICSTRIM, currentTrimValue>>1);     //  8-bits
      if ((prefix(target_sync)() != 0) && (prefix(target_sync)() != 0)) {
         fprintf(stderr, "Failed to re-connect to Target\n");
         return(0);
      }
      currentSpeed =prefix(get_speed)();
      printf(" Trim values %2X,%2X, Freq=%f", currentTrimValue>>1, currentTrimValue&0x01, currentSpeed);
      printf("%s %f\n", (currentSpeed < targetFrequency)?"<":">=", targetFrequency);
      if (currentSpeed < targetFrequency){ // Too slow - remove trial bit
         currentTrimValue &= ~trialBit;
      }
   }

   // Assume binary search value is "best" so far
   int   bestValue      = currentTrimValue;
   int   limitTrimValue = 0;
   float bestFreq       = currentSpeed;
   float bestError      = fabs(targetFrequency-currentSpeed);
   float currentError   = 0;

   printf("2. Linear search\n");

   // Linear search +/-SEARCH_OFFSET, starting at higher freq (smaller Trim)
   // Range is constrained to [1..254]
   if (currentTrimValue >= (254-SEARCH_OFFSET))
      limitTrimValue = 254;
   else
      limitTrimValue = currentTrimValue + SEARCH_OFFSET;
   if (currentTrimValue <= 1+SEARCH_OFFSET)
      currentTrimValue = 1;
   else
      currentTrimValue -= SEARCH_OFFSET;

   while (currentTrimValue <= limitTrimValue) {
      prefix(write_byte)(addrICSSC,   currentTrimValue&0x01);   //  1-bit
      prefix(write_byte)(addrICSTRIM, currentTrimValue>>1);     //  8-bits
      if ((prefix(target_sync)() != 0) && (prefix(target_sync)() != 0)) {
         fprintf(stderr, "Failed to re-connect to Target\n");
         return(0);
      }
      currentSpeed = prefix(get_speed)();
      currentError = fabs(targetFrequency-currentSpeed);

      if (currentError < bestError){ // Better Guess?
         printf(" Trim values* %2X,%2X, Freq=%f", currentTrimValue>>1, currentTrimValue&0x01, currentSpeed);
         printf("err= %f\n", currentError);
         bestValue    = currentTrimValue;
         bestFreq     = currentSpeed;
         bestError    = currentError;
      }
      else {
         printf(" Trim values  %2X,%2X, Freq=%f", currentTrimValue>>1, currentTrimValue&0x01, currentSpeed);
         printf("err= %f\n", currentError);
      }
      currentTrimValue++;
   }

   printf("=Trim values ICSTRM=%3X, ICSSC=%3X, Freq=%f\n", bestValue>>1, bestValue&0x01, bestFreq);
   printf("=Trim values ICSTRM=%3d, ICSSC=%3d, Freq=%f\n", bestValue>>1, bestValue&0x01, bestFreq);

   //trim <mode=8/9 bit> <ICSTRM> <ICSSC> <NV_ICSTRM> <NV_FTRIM> <Target_Frequency>
   printf(":trim %d %4.4X %4.4X %4.4X %4.4X %fn", mode, addrICSTRIM, addrICSSC, addrNV_ICSTRM, addrNV_FTRIM, targetFrequency);
   return 0;
}

//! Write Control Register
int writeControlCommand(void) {
// wc <control_value>
   char *currentToken;
   int value;

   // Read control value to set
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%X",&value) != 1))
      return 1;

   prefix(write_control)((U8) value);
   printf(":wc %2.2X\n", value);
   return 0;
}

//! Set communication speed
int setSpeedCommand(void) {
// wc <control_value>
   char *currentToken;
   float speed;
   int rc;

   // Read speed value to set
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%f",&speed) != 1))
      return 1;

   rc = prefix(set_speed)(speed);
   if (rc != 0)
      printf("Setting speed failed\n");

   printf("Speed = %.2f MHz (%.0f ticks, sync=%.1f us)\n",
          speed, (60.0 * 2 * 128)/speed, (2 * 128)/speed);
   return 0;
}

//! Set target Type
int setTargetCommand(void) {
// setTarget <value>
   char *currentToken;
   int   value;

   // Read control value to set
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%d",&value) != 1))
      return 1;

   targetType = (U8) value;

   if (prefix(set_target_type)((U8)value) != 0)
      printf(":setTarget %d - Failed\n", value);
   else
      printf(":setTarget %d\n", value);
   return 0;
}

//! Set Target to re-boot into ICP mode
int setBootCommand(void) {
// setBoot
   prefix(icp_reboot)();

   printf("setBootCommand Complete\n");
   return 0;
}

//! Send debug command to BDM
int debugCommand(void) {
// debug <control_value>
   char *currentToken;
   int value;
   unsigned char usb_data[20] = {0};

   // Read control value to set
   currentToken = strtok(NULL, delimiters);
   if ((currentToken==NULL) || (sscanf(currentToken,"%d",&value) != 1))
      return 1;

   usb_data[0]=2;  /* receive 2 bytes */
   usb_data[1]=CMD_DEBUG;
   usb_data[2]=value;

   bdm_debug(usb_data);

   printf(":debug %2.2X => %2.2X(%2.2d), %2.2X(%2.2d), "
          "%2.2X(%2.2d), %2.2X(%2.2d) \n", value,
          usb_data[0], usb_data[0], usb_data[1], usb_data[1],
          usb_data[2], usb_data[2], usb_data[3], usb_data[3] );
   reportState();

   return 0;
}

//! Exit program
int exitCommand(void) {
   return 1;
}

//! Do nothing
int commentCommand( void ) {
   return 0;
}

//! Usage message
static const char usageText[] =
   "connect                               - connect to target\n\r"
   "debug <value>                         - debug commands\n\r"
   "delay <time>                          - pause a while\n\r"
   "db <addr>                             - read byte\n\r"
   "dw <addr>                             - read word\n\r"
   "dl <addr>                             - read longword (CFV1 only)\n\r"
   "exit                                  - exit program\n\r"
   "go                                    - start from current PC\n\r"
   "gs                                    - read status register\n\r"
   "halt                                  - halt the target\n\r"
   "help                                  - display help\n\r"
   "regs                                  - print out registers\n\r"
   "reset                                 - Reset to bdm mode\n\r"
   "resetn                                - Reset to normal mode\n\r"
   "rb <addr>                             - read byte\n\r"
   "rblock <addr> <size>                  - read block\n\r"
   "rw <addr>                             - read word\n\r"
   "rl <addr>                             - read longword (CFV1 only)\n\r"
   "setboot                               - set USBDM module to ICP mode\n\r"
   "settarget <target>                    - 0=HCS12, 1=HCS08, 2=RS08, 3=CFV1\n\r"
   "speed <Hz>                            - set speed \n\r"
   "step                                  - execute a single instruction\n\r"
   "tblock <start> <end> <repeatcount>    - random RAM write/read block test\n\r"
   "trim <params>                         - calculate clock trim value\n\r"
   "wait <time>                           - pause a while\n\r"
   "wb <addr> <data>                      - write byte\n\r"
   "wblock <addr> <size> <data>           - write block\n\r"
   "wc <value>                            - write control\n\r"
   "wpc <value>                           - write to PC\n\r"
   "ww <addr><value>                      - write word\n\r"
   "wl <addr><value>                      - write longword (CFV1 only)\n\r"
   ;

//! Help command
int helpCommand(void) {
   fprintf(stderr,"%s",usageText);
   return 0;
}

//! Command Entry structure
typedef struct {
   char *command;       //!< Command name (string)
   int (*func)(void);   //!< Command function to execute
} CommandEntry;

//! Command Table
CommandEntry commandTable[] = {
   {"connect",   connectCommand},
   {"debug",     debugCommand},
   {"delay",     delayCommand},
   {"db",        rbCommand},
   {"dw",        rwCommand},
   {"dl",        rlCommand},
   {"exit",      exitCommand},
   {"go",        goCommand},
   {"gs",        getStatusCommand},
   {"halt",      haltCommand},
   {"help",      helpCommand},
   {"regs",      getRegsCommand},
   {"reset",     resetSpecialCommand},
   {"resetn",    resetNormalCommand},
   {"rb",        rbCommand},
   {"rblock",    rblockCommand},
   {"rw",        rwCommand},
   {"rl",        rlCommand},
   {"setboot",   setBootCommand},
   {"settarget", setTargetCommand},
   {"speed",     setSpeedCommand},
   {"step",      stepCommand},
   {"tblock",    testBlockCommand},
   {"trim",      trimCommand},
   {"wait",      delayCommand},
   {"wb",        wbCommand},
   {"wblock",    wblockCommand},
   {"wc",        writeControlCommand},
   {"wpc",       wpcCommand},
   {"ww",        wwCommand},
   {"wl",        wlCommand},
   {"?",         helpCommand},
   {"wreg",      wRegCommand},
   {"wcreg",     wCregCommand},
   {"wdreg",     wDregCommand},
   {"rreg",      rRegCommand},
   {"rcreg",     rCregCommand},
   {"rdreg",     rDregCommand},
   {"testcreg",  testCregCommand},
};

//! Search for & execute command
//!
int execute(char *commandLine){
   char           *currentToken;
   CommandEntry   *commandEntry;

   // Remove comments
   currentToken = strstr(commandLine, "//");
   if (currentToken != NULL){
      *currentToken = '\0';
   }
   currentToken = strtok(commandLine, delimiters);

   if (currentToken == NULL)
      return 0;

   for( commandEntry = commandTable;
        commandEntry < commandTable+sizeof(commandTable)/sizeof(commandTable[0]);
        commandEntry++ ){
      if (strcmp(currentToken,commandEntry->command)==0){
         return commandEntry->func();
      }
   }
   printf("Unknown command \r\n");
   return 0;
}

//! Mainline
int main(void){
int numDevices;
char commandBuffer[100];
int errCode;

   HINSTANCE hInst = GetModuleHandle(0);

   dll_initialize(hInst);

   numDevices = prefix(init)();

   if (numDevices == 0) {
      fprintf(stderr, "No devices Found\n");
      return 1;
   }
   prefix(open)(0);

   reportState();

   while (gets(commandBuffer) != NULL){
      printf("Executing:%s\n", commandBuffer);
      errCode = execute(commandBuffer);
      if ((errCode != 0)) {
         printf("Failed\n");
         break;
      }
   }
   prefix(close)();
   dll_uninitialize();
   return 0;
}

#if 0
int main2(void) {
   char commandBuffer[100];
   FILE *commandFile;
   int errCode;

   commandFile = fopen("MC9S08QG8.trim", "rt");
   if (commandFile == NULL){
      fprintf(stderr,"Unable to open trim file");
      return 0;
   }
   dev = OSBDMConnect();
   if (dev == NULL) {
      printf("error: device not found!\n");
      return 0;
   }

   while (gets(commandBuffer/*, 80, commandFile*/) != NULL){
      printf("Executing:%s\n", commandBuffer);
      errCode = execute(commandBuffer);
      if ((errCode != 0)) {
         printf("Failed\n");
         break;
      }
   }

   //========================================================
   //CLEANUP:
   if (dev != NULL) {
      usb_release_interface(dev, 0);
      usb_close(dev);
   }

   if (commandFile != NULL)
      fclose(commandFile);

  return 0;
}

#endif
