/*! \file
    \brief Header file for OSBDM routines

*/
#ifndef _USBDM_H_
#define _USBDM_H_
//__stdcall
// __cdecl

//! This macro adds the correct prefix to a DLL entry point name
#if defined(TBLCF)
#define prefix(x) _tblcf_ ## x
#define API WINAPI
#elif defined(TBDML)
#define prefix(x) _tbdml_ ## x
#define API WINAPI
#else // default to OSBDM
#define prefix(x) _opensourcebdm_ ## x
#define API WINAPI
#endif

#if defined(DLL)
//! This definition is used when OSBDM.h is being exported (creating DLL)
#define USBDM_API API __declspec(dllexport)
#elif defined (COMMANDLINE) || defined (DOC)
//! This empty definition is used for the command-line version that statically links the DLL routines
#define USBDM_API
#else
//! This definition is used when USBDM.h is being imported (dynamically linked)
#define USBDM_API API __declspec(dllimport)
#endif

#define WRITE_BLOCK_CHECK //<! Enable checking of send only USB commands commands

//! MACRO to calculate a SYNC value from a given frequency in Hz
#define SYNC_MULTIPLE(x)  (U16)((2*128*(60000000UL/10))/(x/10))

//===================================================================================
//!  Target microcontroller types
typedef enum {
   T_HC12      = 0,     //!< HC12 or HCS12 target
   T_HCS08     = 1,     //!< HCS08 target
   T_RS08      = 2,     //!< RS08 target
   T_CFV1      = 3,     //!< Coldfire Version 1 target
   T_CFVx      = 4,     //!< Coldfire Version 2,3,4 target
   T_JTAG      = 5,     //!< JTAG target - TAP is set to \b RUN-TEST/IDLE
   T_EZFLASH   = 6,     //!< EzPort Flash interface (SPI?)
   T_OFF       = 0xFF,  //!< Turn off interface (no target)
} target_type_e;

//!  Reset mode (normal operation or BDM)
typedef enum {	/* type of reset action required */
	SPECIAL_MODE = 0,  //!< Special mode (BDM active, Target halted)
	NORMAL_MODE  = 1   //!< Normal mode (usual reset, Target executes)
} target_mode_e;

//!  Target RS08 microcontroller derivatives
typedef enum {
  KA1 = 0, //!< RS08KA1
  KA2 = 1, //!< RS08KA2
  LA8 = 2, //!< RS08LA8
  KA4 = 4, //!< RS08KA4 (testing only)
  KA8 = 5, //!< RS08KA8 (testing only)
} rs08_derivative_type_e;

//! Target supports ACKN or use fixed delay (WAIT) instead
typedef enum {
   ACKN  = 1,   //!< Target supports ACKN feature and it is enabled
   WAIT  = 0,   //!< Use WAIT (delay) instead
} ackn_e;

//! Target reset status values
typedef enum {
   NO_RESET_ACTIVITY    = 0,   //!< No reset activity since last polled
   RESET_DETECTED       = 1    //!< Reset since last polled
} reset_e;

//! Target RSTO state
typedef enum {
	RSTO_ACTIVE=0,     //!< RSTO is currently low (active)
	RSTO_INACTIVE=1    //!< RSTO is currently high (inactive)
} reset_state_e;

//! Target speed selection
typedef enum {
   NO_INFO              = 0,   //!< Not connected
   SYNC_SUPPORTED       = 1,   //!< Speed determined by SYNC
   SPEED_GUESSED        = 2,   //!< Speed determined by trial & error
   SPEED_USER_SUPPLIED  = 3    //!< User has specified the speed to use
} speed_e;

//! (HCS12/HCS08/RS08/Coldfire V1)  State of BDM Communication
typedef struct {
   ackn_e         ackn_state;       //!< Supports ACKN ?
   reset_e        reset_state;      //!< Target reset recently ?
   speed_e        connection_state; //!< Connection status & speed determination method
} bdm_status_t;

//! (Coldfire V2, V3 and V4)  State of BDM Communication
typedef struct {
	reset_state_e   reset_state;       //!<  Current target RST0 state
	reset_e         reset_detection;   //!<  Target reset recently?
} bdmcf_status_t;

//! Target microcontroller registers (HCS12, HCS08/RS08)
typedef union {
   //! HCS12 Registers
	struct {
		unsigned int pc;  //!< Program counter
		unsigned int sp;  //!< Stack Pointer
		unsigned int ix;  //!< Index register X
		unsigned int iy;  //!< Index register Y
		unsigned int d;   //!< D accumulator (A:B)
		unsigned int ccr; //!< Condition code register
	} hc12;
	//! HCS08 Registers (also used for RS08)
	struct {
		unsigned int pc;  //!< Program Counter (HCS08), CCR+PC (RS08)
		unsigned int sp;  //!< Stack Pointer (HCS08), Shadow PC (RS08)
		unsigned int hx;  //!< H:X index register (HCS08 only)
		unsigned int a;   //!< Accumulator (HCS08 + RS08)
		unsigned int ccr; //!< Condition code register (HCS08 only)
	} hcs08;
} registers_t;

#ifdef TBLCF
USBDM_API unsigned int  prefix(DLL_VERSION)(void);                            /* returns DLL version in BCD */
#else
USBDM_API unsigned char  prefix(dll_version)(void);                           /* returns DLL version in BCD */
#endif
USBDM_API unsigned char  prefix(init)(void);						                  /* initialises USB and returns number of devices found */
USBDM_API unsigned char  prefix(init_hardware)(void);
USBDM_API unsigned char  prefix(open)(unsigned char device_no); 	            /* opens a device with given number (0...), returns 0 on success and 1 on error */
USBDM_API void           prefix(close)(void);								         /* closes currently open device */
USBDM_API unsigned int   prefix(get_version)(void);				               /* returns version of HW (MSB) and SW (LSB) of the interface in BCD format */
USBDM_API unsigned char  prefix(get_last_sts)(void);				               /* returns status of the last command: 0 on success and non-zero on failure */
USBDM_API unsigned char  prefix(set_target_type)(target_type_e target_type);  /* sets target MCU type, returns 0 on success and non-zero on failure */

USBDM_API unsigned char  prefix(target_go)(void);                             /* starts target execution from current PC address, returns 0 on success and non-zero on failure */
USBDM_API unsigned char  prefix(target_step)(void); 				               /* steps over a single target instruction, returns 0 on success and non-zero on failure */
USBDM_API unsigned char  prefix(target_halt)(void); 				               /* brings the target into active background mode, returns 0 on success and non-zero on failure */
USBDM_API unsigned char  prefix(target_reset)(target_mode_e target_mode);     /* resets the target to normal or special mode, returns 0 on success and non-zero on failure */

#if defined(TBDML) || defined(OSBDM)
USBDM_API unsigned char  prefix(target_sync)(void); 				            /* measures BDM frequency of the target using the SYNC feature and connects to it, returns 0 on success and non-zero on failure */
USBDM_API unsigned char  prefix(bdm_sts)(bdm_status_t *bdm_status);           /* fills user supplied structure with actual state of BDM communication channel, returns 0 on success and non-zero on failure */

USBDM_API unsigned char  prefix(read_bd)(unsigned int address); 	            /* reads one byte from the BDM area of memory */
USBDM_API unsigned char  prefix(write_bd)(unsigned int address, unsigned char data); /* writes one byte to the BDM area of memory, returns 0 on success and non-zero on failure */
USBDM_API unsigned char  prefix(set_speed)(float crystal_frequency);          /* sets the communication speed, parameter is crystal frequency in MHz, returns 0 on success and non-zero on failure */
USBDM_API float          prefix(get_speed)(void); 						            /* returns crystal frequency of the target in MHz */
USBDM_API unsigned char  prefix(read_byte)(unsigned int address);             /* reads one byte from memory */
USBDM_API void           prefix(write_byte)(unsigned int address, unsigned char data); /* writes one byte to memory */
USBDM_API unsigned int   prefix(read_word)(unsigned int address); 	         /* reads one word from memory (address must be aligned) */
USBDM_API void           prefix(write_word)(unsigned int address, unsigned int data); /* writes one word to memory (address must be aligned) */
USBDM_API void           prefix(read_block)(unsigned int address, unsigned int count, unsigned char *data); /* reads block from memory, data is returned to the user supplied buffer */
USBDM_API void           prefix(write_block)(unsigned int address, unsigned int count, unsigned char *data); /* writes block to memory */
USBDM_API unsigned char  prefix(read_regs)(registers_t *registers);           /* reads contents of core registers, returns 0 on success and non-zero on failure */
USBDM_API void           prefix(write_reg_pc)(unsigned int value); 		      /* writes PC */
USBDM_API void           prefix(write_reg_sp)(unsigned int value); 		      /* writes SP */
USBDM_API void           prefix(write_reg_x)(unsigned int value);	 	         /* writes H:X (HCS08) or IX (HC12) */
USBDM_API void           prefix(write_reg_y)(unsigned int value); 		      /* writes IY (HC12 only) */
USBDM_API void           prefix(write_reg_d)(unsigned int value); 		      /* writes B:A (HC12) or A (HCS08) */
USBDM_API void           prefix(write_reg_ccr)(unsigned int value); 	         /* writes CCR */
#endif // defined(TBDML) || defined(OSBDM)

#ifdef OSBDM
USBDM_API void           prefix(mem_dlstart)(void);
USBDM_API void           prefix(mem_dlend)(void);
USBDM_API unsigned char  prefix(target_readyfor_datald)(void);
USBDM_API unsigned int   prefix(read_bkpt)(void);
USBDM_API void           prefix(write_bkpt)(unsigned int value);
USBDM_API unsigned char  prefix(read_status)(void);
USBDM_API void           prefix(write_control)(unsigned char value);
USBDM_API unsigned char  prefix(set_derivative_type)(unsigned char derivative);

USBDM_API unsigned int   prefix(read_xcsr_byte)(void);
USBDM_API unsigned int   prefix(read_reg)(unsigned char regNo);
USBDM_API void           prefix(write_reg)(unsigned char regNo,   unsigned int value);
USBDM_API unsigned int   prefix(read_creg)(unsigned char regNo);
USBDM_API void           prefix(write_creg)(unsigned char regNo,  unsigned int value);
USBDM_API unsigned int   prefix(read_dreg)(unsigned char regNo);
USBDM_API void           prefix(write_dreg)(unsigned char regNo,  unsigned int value);
USBDM_API BOOL           prefix(read_mem)(unsigned int addr,      unsigned int size,
                                          unsigned int memSpace,  unsigned char *data);
USBDM_API BOOL           prefix(write_mem)(unsigned int addr,     unsigned int count,
                                           unsigned int memSpace, unsigned char *data);
#endif

#ifdef TBLCF
USBDM_API unsigned char  prefix(resynchronize)(void); 			               /* resynchronizes communication with the target (in case of noise, etc.); returns 0 on success and non-zero on failure */
USBDM_API unsigned char  prefix(bdm_sts)(bdmcf_status_t *bdmcf_status); 		/* fills user supplied structure with current state of the BDM communication channel; returns 0 on success and non-zero on failure */
USBDM_API unsigned char  prefix(assert_ta)(unsigned char duration_10us); 		// asserts the TA signal for the specified time (in 10us ticks); returns 0 on success and non-zero on failure */
USBDM_API unsigned char  prefix(read_reg)(unsigned char reg_index, unsigned long int * result);	   /* reads the specified register and writes its contents into the supplied buffer; returns 0 on success and non-zero on failure */
USBDM_API void           prefix(write_reg)(unsigned char reg_index, unsigned long int value); 		/* writes specified register */
USBDM_API unsigned char  prefix(read_creg)(unsigned int address, unsigned long int * result);		/* reads control register at the specified address and writes its contents into the supplied buffer; returns 0 on success and non-zero on failure */
USBDM_API void           prefix(write_creg)(unsigned int address, unsigned long int value);			/* writes control register at the specified address; returns 0 on success and non-zero on failure */
USBDM_API unsigned char  prefix(read_dreg)(unsigned char dreg_index, unsigned long int * result);  /* reads the specified debug register and writes its contents into the supplied buffer; returns 0 on success and non-zero on failure */
USBDM_API void           prefix(write_dreg)(unsigned char dreg_index, unsigned long int value); 	/* writes specified debug register */
USBDM_API unsigned char  prefix(read_mem8)(unsigned long int address, unsigned char * result); 	   /* reads byte from the specified address; returns 0 on success and non-zero on failure */
USBDM_API unsigned char  prefix(read_mem16)(unsigned long int address, unsigned int * result);	   /* reads word from the specified address; returns 0 on success and non-zero on failure */
USBDM_API unsigned char  prefix(read_mem32)(unsigned long int address, unsigned long int * result);/* reads long word from the specified address; returns 0 on success and non-zero on failure */
USBDM_API void           prefix(write_mem8)(unsigned long int address, unsigned char value);			/* writes byte at the specified address */
USBDM_API void           prefix(write_mem16)(unsigned long int address, unsigned int value); 		/* writes word at the specified address */
USBDM_API void           prefix(write_mem32)(unsigned long int address, unsigned long int value);	/* writes long word at the specified address */
USBDM_API unsigned char  prefix(read_block8)(unsigned long int address, unsigned long int bytecount, unsigned char *buffer);		/* reads the requested number of bytes from target memory from the supplied address and stores results into the user supplied buffer; uses byte accesses only; returns 0 on success and non-zero on failure */
USBDM_API unsigned char  prefix(read_block16)(unsigned long int address, unsigned long int bytecount, unsigned char *buffer); 	/* reads the requested number of bytes from target memory from the supplied address and stores results into the user supplied buffer; uses word accesses; returns 0 on success and non-zero on failure */
USBDM_API unsigned char  prefix(read_block32)(unsigned long int address, unsigned long int bytecount, unsigned char *buffer);		/* reads the requested number of bytes from target memory from the supplied address and stores results into the user supplied buffer; uses long word accesses; returns 0 on success and non-zero on failure */
USBDM_API unsigned char  prefix(write_block8)(unsigned long int address, unsigned long int bytecount, unsigned char *buffer); 	/* writes the requested number of bytes to target memory from the supplied address; uses byte accesses only; returns 0 on success and non-zero on failure (must be compiled with WRITE_BLOCK_CHECK, otherwise always returns 0) */
USBDM_API unsigned char  prefix(write_block16)(unsigned long int address, unsigned long int bytecount, unsigned char *buffer); 	/* writes the requested number of bytes to target memory at the supplied address; uses word accesses; returns 0 on success and non-zero on failure (must be compiled with WRITE_BLOCK_CHECK, otherwise always returns 0) */
USBDM_API unsigned char  prefix(write_block32)(unsigned long int address, unsigned long int bytecount, unsigned char *buffer); 	/* writes the requested number of bytes to target memory at the supplied address; uses long word accesses; returns 0 on success and non-zero on failure (must be compiled with WRITE_BLOCK_CHECK, otherwise always returns 0) */

USBDM_API unsigned char  prefix(jtag_sel_shift)(unsigned char mode);								/* JTAG - go from RUN-TEST/IDLE to SHIFT-DR (mode==0) or SHIFT-IR (mode!=0) state */
USBDM_API unsigned char  prefix(jtag_sel_reset)(void);												/* JTAG - go from RUN-TEST/IDLE to TEST-LOGIC-RESET state */
USBDM_API void           prefix(jtag_write)(unsigned char bit_count, unsigned char exit, unsigned char *buffer);	/* JTAG - write data; parameter exit: ==0 : stay in SHIFT-xx, !=0 : go to RUN-TEST/IDLE when done; data: shifted in LSB (last byte) first, unused bits (if any) are in the MSB (first) byte */
USBDM_API unsigned char  prefix(jtag_read)(unsigned char bit_count, unsigned char exit, unsigned char *buffer);	/* JTAG - read data; parameter exit: ==0 : stay in SHIFT-xx, !=0 : go to RUN-TEST/IDLE when done; data: shifted in LSB (last byte) first, unused bits (if any) are in the MSB (first) byte */
#endif // TBLCF

USBDM_API void           prefix(setBoot)(void);

USBDM_API void           prefix(icp_reboot)( void );
USBDM_API unsigned int   prefix(icp_program)(unsigned int addr, unsigned int count, unsigned char *data);
USBDM_API unsigned int   prefix(icp_verify)(unsigned int addr, unsigned int count, unsigned char *data);
USBDM_API unsigned int   prefix(icp_erase)(unsigned int addr, unsigned int count);
USBDM_API unsigned int   prefix(icp_get_result)( void );

unsigned int bdm_set_option(unsigned char option, unsigned int value);

#if defined (COMMANDLINE)
void bdm_debug(char usb_data[]);
void dll_initialize(HINSTANCE);
void dll_uninitialize(void);
#endif

#endif // _USBDM_H_
