/*! \file
   \brief Entry points for opensourcebdm.dll library

   \verbatim
   USBDM interface DLL
   Copyright (C) 2008  Peter O'Donoghue

   Based on
      Opensource BDM  - interface DLL

   which is based on
      TBDML interface DLL by
      Copyright (C) 2005  Daniel Malik

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
\endverbatim

\verbatim
 Change History
+==============================================================================
|  1 Jan 2009 | Changed to common file for TBDML, OSBDM & TBLCF dlls      - pgo
| 27 Dec 2008 | Added extendedConnect() to target_sync()                  - pgo
| 21 Oct 2008 | Increased size of address in ICP commands                 - pgo
| 20 Oct 2008 | Fixed size of SET_BOOT command                            - pgo
+==============================================================================
\endverbatim
*/
#include <windows.h>
#include <stdio.h>
#include "log.h"
#include "version.h"
#include "interface.h"
#include "Common.h"
#include "ICP.h"
#include "common_usb.h"
#include "commands.h"
#include "Configure.h"
#include "CmdTable.h"

/*
 * OSBDM_CF
   Index: 1    Name: osbdm_DLL_VERSION       x
   Index: 2    Name: osbdm_close             x
   Index: 3    Name: osbdm_get_version       x
   Index: 4    Name: osbdm_init              x
   Index: 5    Name: osbdm_init_hardware     x
   Index: 6    Name: osbdm_open              x
   Index: 7    Name: osbdm_read16
   Index: 8    Name: osbdm_read32
   Index: 9    Name: osbdm_read_bd
   Index: 10   Name: osbdm_read_block
   Index: 11   Name: osbdm_send1
   Index: 12   Name: osbdm_status            x
   Index: 13   Name: osbdm_target_go         x
   Index: 14   Name: osbdm_target_halt       x
   Index: 15   Name: osbdm_target_reset      x
   Index: 16   Name: osbdm_target_step       x
   Index: 17   Name: osbdm_write16
   Index: 18   Name: osbdm_write32
   Index: 19   Name: osbdm_write_bd
   Index: 20   Name: osbdm_write_block
   Index: 21   Name: osbdm_write_fill
 *
 */
#if 0
USBDM_API unsigned char prefix(status)(void) {
   return 0;
}
#endif

/*
 *  Define symbol "LOG" in "log.h" during compilation to produce a log file tbdml_dll.log
 */

static unsigned char usb_data[MAX_PACKET_SIZE+1]; //!< Buffer for USB I/O messages
static HINSTANCE hDLLInst = NULL;                 //!< Handle on DLL for resources

//! Get version of the DLL in BCD format
//!
//! @return 8-bit version number (2 BCD digits, major-minor)
//!
#ifdef TBLCF
USBDM_API unsigned int  prefix(DLL_VERSION)(void){
#else
USBDM_API unsigned char  prefix(dll_version)(void){
#endif
   print("dll_version() => %2.2D\r\n", USBDM_DLL_VERSION);
	return (USBDM_DLL_VERSION);
}

//! Initialises USB interface
//!
//! This must be done before any device may be opened.
//!
//! @return Number of BDM devices found
//!
USBDM_API unsigned char prefix(init)(void) {
	unsigned char i;

	bdm_usb_find_devices();	/* look for devices on all USB busses */
	i = bdm_usb_cnt();
	print("init(): Usb initialised, found %d device(s)\r\n",i);
	return (i);					/* count of devices found */
}
#if 0
//! Initialises USB hardware
//!
//! @return ??
//!
USBDM_API unsigned char prefix(init_hardware)(void) {

   print("init_hardware()\r\n");
   return 1;
}
#endif

//! Opens a device
//!
//! @param device_no Number (0..N) of device to open.
//! A device must be open before any communication with the device can take place.
//! @note The range of device numbers must be obtained from init() before calling
//! this function.
//! @return 0 => Success,\n !=0 => Fail
//!
USBDM_API unsigned char prefix(open)(unsigned char device_no) {
int rc = 0;

   print("open(): Trying to open device #%d\r\n",device_no);
   rc = bdm_usb_open(device_no);
   print("open(): rc = #%d\r\n", rc);
   if ((rc == 0) & (bdm_usb_cnt() == 1)) {
     // Successful open and only one device available
     checkFirmware(hDLLInst); // Check Firmware version
   }
   return (rc);
}

//! Closes currently open device
//!
USBDM_API void prefix(close)(void) {
	print("close(): Trying to close the device\r\n");

	// Tell the BDM that we're finished (power down)
   usb_data[0]=2;  /* receive 2 byte */
   usb_data[1]=CMD_SET_TARGET;
   usb_data[2]=T_OFF;
   bdm_usb_recv_ep0(usb_data);
	bdm_usb_close();
}

//! Gets software version and type of hardware
//!
//! @return 32-bit version number (4 bytes)
//!  - [0] = BDM software version                \n
//!  - [1] = hardware type                       \n
//!  - [2] = ICP software version                \n
//!  - [3] = hardware type (from ICP flash area) \n
//!  Bytes [1] and [3] should always agree
//!
USBDM_API unsigned int  prefix(get_version)(void) {
   print("get_version()\r\n");
   usb_data[0] = 5;              // receive up to 5 bytes
	usb_data[1] = CMD_GET_VER;
	bdm_usb_recv_ep0(usb_data);
   print("get_version()=>%4.4X\r\n", (*((U32 *)(usb_data+1))));
	return (*((U32 *)(usb_data+1)));
}

//! Get status of the last command
//!
//! @return 0 => Success,\n 1 => Failure
//!
USBDM_API unsigned char prefix(get_last_sts)(void) {
   print("get_last_sts()\r\n");
   usb_data[0] = 1;              // receive up to 1 byte
	usb_data[1] = CMD_GET_LAST_STATUS;
	bdm_usb_recv_ep0(usb_data);
   print("get_last_sts()=>%d\r\n", usb_data[0]);
	return ((usb_data[0]==BDM_RC_FAIL)||(usb_data[0]==BDM_RC_ILLEGAL_COMMAND));
}

//! Sets target MCU type
//!
//! @param target_type type of target, see \ref target_type_e
//!
//! @return 0 => Success,\n !=0 => Fail
//!
//! @note A dialog box will open to allow the user to configure the BDM
//!    The user may be prompted to supply power to target
//!
USBDM_API unsigned char prefix(set_target_type)(target_type_e target_type) {
   print("set_target_type()\r\n");
#ifdef TBLCF
   target_type = T_CFVx;
#endif
   configureDialog(hDLLInst, target_type); // Do configuration dialogue

   return set_target_type_with_retry(target_type);
}

//======================================================================================
// connect() and resynchronize() serve similar purposes
//
#if defined(TBDML) || defined(OSBDM)
extern unsigned char extendedConnect(HINSTANCE hDLLInst);

//! Connects to Target.
//!
//! This will cause the BDM module to attempt to connect to the Target.
//! In most cases the BDM module will automatically determine the connection
//! speed and successfully connect.  If unsuccessful, it may be necessary
//! to manually set the speed using set_speed().
//! @note If there are connection problems, the user may be prompted to cycle the target power on some targets.
//!
//! @return 0 => Success,\n !=0 => Fail
//!
USBDM_API unsigned char prefix(target_sync)(void) {
   return extendedConnect(hDLLInst);
}
//! Resets the target to normal or special mode
//!
//! @param target_mode see \ref target_mode_e
//!
//! @return 0 => Success,\n !=0 => Fail
//!
USBDM_API unsigned char prefix(target_reset)(target_mode_e target_mode) {
   print("target_reset()\r\n");
   usb_data[0] = 1;  // receive up to 1 byte
   usb_data[1] = CMD_RESET;
   usb_data[2] = target_mode;
   bdm_usb_recv_ep0(usb_data);
   return (usb_data[0]!=CMD_RESET);
}

//! Starts target execution from current PC address
//!
//! @return 0 => Success,\n !=0 => Fail
//!
USBDM_API unsigned char prefix(target_go)(void) {
   print("target_go()\r\n");
   usb_data[0] = 1;  // receive up to 1 byte
   usb_data[1] = CMD_GO;
   bdm_usb_recv_ep0(usb_data);
   return (usb_data[0]!=CMD_GO);
}

//! Steps over a single target instruction
//!
//! @return 0 => Success,\n !=0 => Fail
//!
USBDM_API unsigned char prefix(target_step)(void) {
   print("target_step()\r\n");
   usb_data[0] = 1;  // receive up to 1 byte
   usb_data[1] = CMD_STEP;
   bdm_usb_recv_ep0(usb_data);
   return (usb_data[0]!=CMD_STEP);
}

//! Brings the target into active background mode.  The target will be halted.
//!
//! @return 0 => Success,\n !=0 => Fail
//!
USBDM_API unsigned char prefix(target_halt)(void) {
   print("target_halt()\r\n");
   usb_data[0] = 1;  // receive up to 1 byte
   usb_data[1] = CMD_HALT;
   bdm_usb_recv_ep0(usb_data);
   return (usb_data[0]!=CMD_HALT);
}

#elif defined(TBLCF)

//! Resynchronises communication with the target (in case of noise, etc.) */
//!
//! @return 0 => Success,\n !=0 => Fail
USBDM_API unsigned char prefix(resynchronize)(void) {
	usb_data[0]=1;	 /* get 1 byte */
	usb_data[1]=CMD_CFVx_RESYNCHRONIZE;
	bdm_usb_recv_ep0(usb_data);
	print("resynchronize(): (0x%02X)\r\n",usb_data[0]);
	return(usb_data[0]!=CMD_CFVx_RESYNCHRONIZE);
}

//! Resets the target to normal or special mode
//!
//! @param target_mode see \ref target_mode_e
//!
//! @return 0 => Success,\n !=0 => Fail
//!
USBDM_API unsigned char prefix(target_reset)(target_mode_e target_mode) {
   print("target_reset()\r\n");
   usb_data[0] = 1;  // receive up to 1 byte
   usb_data[1] = CMD_CFVx_RESET;
   usb_data[2] = target_mode;
   bdm_usb_recv_ep0(usb_data);
   return (usb_data[0]!=CMD_CFVx_RESET);
}

//! Starts target execution from current PC address
//!
//! @return 0 => Success,\n !=0 => Fail
//!
USBDM_API unsigned char prefix(target_go)(void) {
   print("target_go()\r\n");
   usb_data[0] = 1;  // receive up to 1 byte
   usb_data[1] = CMD_CFVx_GO;
   bdm_usb_recv_ep0(usb_data);
   return (usb_data[0]!=CMD_CFVx_GO);
}

//! Steps over a single target instruction
//!
//! @return 0 => Success,\n !=0 => Fail
//!
USBDM_API unsigned char prefix(target_step)(void) {
   print("target_step()\r\n");
   usb_data[0] = 1;  // receive up to 1 byte
   usb_data[1] = CMD_CFVx_STEP;
   bdm_usb_recv_ep0(usb_data);
   return (usb_data[0]!=CMD_CFVx_STEP);
}

//! Brings the target into active background mode.  The target will be halted.
//!
//! @return 0 => Success,\n !=0 => Fail
//!
USBDM_API unsigned char prefix(target_halt)(void) {
   print("target_halt()\r\n");
   usb_data[0] = 1;  // receive up to 1 byte
   usb_data[1] = CMD_CFVx_HALT;
   bdm_usb_recv_ep0(usb_data);
   return (usb_data[0]!=CMD_CFVx_HALT);
}
#endif // TBLCF

//=============================================================================================
// bdm_status() depends on DLL unfortunately
//
#if defined(TBDML) || defined(OSBDM)
//! Fills user supplied structure with state of BDM communication channel
//!
//! @param bdm_status Pointer to structure to receive status, see \ref bdm_status_t
//!
//! @return 0 => Success,\n !=0 => Fail
//!
USBDM_API unsigned char prefix(bdm_sts)(bdm_status_t *bdm_status) {
   unsigned char BDMStatus;
   static const char *connectionStates[] = {"Not Connected", "Sync Speed",
                                            "Guess Speed",   "Manual Speed"};
   usb_data[0] = 3;  // receive up to 2 bytes
	usb_data[1] = CMD_GET_STATUS;
	bdm_usb_recv_ep0(usb_data);
	if (usb_data[0]!=CMD_GET_STATUS)
	   return (1);
   BDMStatus = usb_data[1];

	bdm_status->ackn_state       = (BDMStatus&S_ACKN)?ACKN:WAIT;
	bdm_status->reset_state      = (BDMStatus&S_RESET_DETECT)?RESET_DETECTED:NO_RESET_ACTIVITY;
	switch(BDMStatus&S_COMM_MASK) {
	   case S_NOT_CONNECTED : bdm_status->connection_state = NO_INFO;             break; //!< No connection with target
	   case S_SYNC_DONE     : bdm_status->connection_state = SYNC_SUPPORTED;      break; //!< Target communication speed determined by BDM SYNC
	   case S_GUESS_DONE    : bdm_status->connection_state = SPEED_GUESSED;       break; //!< Target communication speed guessed
	   case S_USER_DONE     : bdm_status->connection_state = SPEED_USER_SUPPLIED; break; //!< Target communication speed specified by user
 	};

   print("Status = %s, %s, %s, %s\r\n", bdm_status->ackn_state?"Ackn":"Wait",
                                        connectionStates[bdm_status->connection_state],
                                        bdm_status->reset_state?"Reset":"No Reset",
                                        (BDMStatus&S_POWER)?"Power":"No Power" );
	return(BDM_RC_OK);
}
#elif defined(TBLCF)
//! Fills user supplied structure with state of BDM communication channel
//!
//! @param bdmcf_status Pointer to structure to receive status, see \ref bdmcf_status_t
//!
//! @return 0 => Success,\n !=0 => Fail
//!
USBDM_API unsigned char prefix(bdm_sts)(bdmcf_status_t *bdmcf_status) {
U16 BDMStatus;

   usb_data[0] = 3;	 /* get 3 bytes */
	usb_data[1] = CMD_GET_STATUS;
	bdm_usb_recv_ep0(usb_data);
	if (usb_data[0]!=CMD_GET_STATUS)
	   return (1);
   BDMStatus = usb_data[1];

   bdmcf_status->reset_state      = (BDMStatus&S_RESET_STATE)?RSTO_INACTIVE:RSTO_ACTIVE;
   bdmcf_status->reset_detection  = (BDMStatus&S_RESET_DETECT)?RESET_DETECTED:NO_RESET_ACTIVITY;

   print("bdm_sts(): Status = %s, %s, %s\r\n", bdmcf_status->reset_state?"RSTO=1":"RSTO=0",
                                               bdmcf_status->reset_detection?"Reset":"No Reset",
                                               (BDMStatus&S_POWER)?"Power":"No Power" );
	return(BDM_RC_OK);
}
#endif

//==================================================================
//  TBLCF dll entry points
//
#if defined(TBLCF)

//! Asserts the TA signal for the specified time
//!
//! @param duration_10us time in (in 10us ticks)
//!
//! @return 0 => OK, != 0 => fail
//!
USBDM_API unsigned char prefix(assert_ta)(unsigned char duration_10us) {
	usb_data[0] = 1;	 /* get 1 byte */
	usb_data[1] = CMD_CFVx_ASSERT_TA;
	usb_data[2] = duration_10us;
	bdm_usb_recv_ep0(usb_data);

	print("assert_ta(): duration %d us (0x%02X)\r\n",(int)10*usb_data[2],usb_data[0]);
	return(usb_data[0] != CMD_CFVx_ASSERT_TA);
}

//! (CF V2, V3, V4) Read control register
//!
//! @param regAddr  Register address
//! @param result   ptr to 32-result buffer
//!
//! @return 0 => OK, != 0 => fail
//!
USBDM_API unsigned char prefix(read_creg)(unsigned int regAddr, unsigned long int *result) {
	usb_data[0] = 5;	 /* get 5 bytes */
	usb_data[1] = CMD_CFVx_READ_CREG;
	usb_data[2] = (regAddr>>8)&0xff;	// Address is big-endian
	usb_data[3] = regAddr&0xff;
	bdm_usb_recv_ep0(usb_data);

	*result = (((unsigned long int)usb_data[1])<<24)+(usb_data[2]<<16)+(usb_data[3]<<8)+usb_data[4];	/* result is in big endian */
   print("read_creg(): %s(0x%04X) == 0x%08lX\r\n",
         getCFV2ControlRegName(regAddr), regAddr, *result);
	return(usb_data[0]!=CMD_CFVx_READ_CREG);
}

//! (CF V2, V3, V4) Write control register
//!
//! @param regAddr  Register address
//! @param value    32-result value to write
//!
//! @return 0 => OK, != 0 => fail
//!
USBDM_API void prefix(write_creg)(unsigned int regAddr, unsigned long int value) {
   usb_data[0] = 7;	 /* send 7 bytes */
	usb_data[1] = CMD_CFVx_WRITE_CREG;
	usb_data[2] = (regAddr>>8)&0xff;  // Address is big-endian
	usb_data[3] = regAddr&0xff;
	usb_data[4] = (value>>24)&0xff;   // Data is big-endian
	usb_data[5] = (value>>16)&0xff;
	usb_data[6] = (value>>8)&0xff;
	usb_data[7] = (value)&0xff;
	print("write_creg(): %s(0x%04X) <= 0x%08lX\r\n",
	      getCFV2ControlRegName(regAddr), regAddr, value);
	bdm_usb_send_ep0(usb_data);
}

//! (CF V2, V3, V4) Reads debug register
//!
//! @param regIndex  Register index
//! @param result    ptr to 32-result buffer
//!
//! @return 0 => OK, != 0 => fail
//!
USBDM_API unsigned char prefix(read_dreg)(unsigned char regIndex, unsigned long int * result) {
	usb_data[0] = 5;	 /* get 5 bytes */
	usb_data[1] = CMD_CFVx_READ_DREG;
	usb_data[2] = regIndex;
	bdm_usb_recv_ep0(usb_data);

	*result = (((unsigned long int)usb_data[1])<<24)+(usb_data[2]<<16)+(usb_data[3]<<8)+usb_data[4];  // Result is big-endian
	print("read_dreg(): Read debug register #0x%02X, result: 0x%08lX (0x%02X)\r\n",regIndex,*result,usb_data[0]);
   print("read_dreg(): %s(0x%04X) == 0x%08lX\r\n",
         getCFDebugRegName(regIndex), regIndex, *result);

	return(usb_data[0]!=CMD_CFVx_READ_DREG);
}

//! (CF V2, V3, V4) Writes debug register
//!
//! @param regIndex  Register index
//! @param value     32-result value to write
//!
//! @return 0 => OK, != 0 => fail
//!
USBDM_API void prefix(write_dreg)(unsigned char regIndex, unsigned long int value) {
	usb_data[0] = 6;	 /* send 6 bytes */
	usb_data[1] = CMD_CFVx_WRITE_DREG;
	usb_data[2] = regIndex;           // Register index
	usb_data[3] = (value>>24)&0xff;   // Value is big-endian
	usb_data[4] = (value>>16)&0xff;
	usb_data[5] = (value>>8)&0xff;
	usb_data[6] = (value)&0xff;
   print("write_dreg(): %s(0x%04X) == 0x%08lX\r\n",
         getCFDebugRegName(regIndex), regIndex, value);

	bdm_usb_send_ep0(usb_data);
}

//! (CF V2, V3, V4) Reads register
//!
//! @param regIndex  Register index
//! @param result    ptr to 32-result buffer
//!
//! @return 0 => OK, != 0 => fail
//!
USBDM_API unsigned char prefix(read_reg)(unsigned char regIndex, unsigned long int *result) {
	usb_data[0]=5;	 /* get 5 bytes */
	usb_data[1]=CMD_CFVx_READ_REG;
	usb_data[2]=regIndex;
	bdm_usb_recv_ep0(usb_data);
	*result = (((unsigned long int)usb_data[1])<<24)+(usb_data[2]<<16)+(usb_data[3]<<8)+usb_data[4];	/* result is in big endian */

	print("read_reg(): %s(0x%04X) == 0x%08lX\r\n",
	                   getCFRegName(regIndex), regIndex, *result);
	return(usb_data[0]!=CMD_CFVx_READ_REG);
}

//! (CF V2, V3, V4) Writes debug register
//!
//! @param regIndex  Register index
//! @param value     32-result value to write
//!
//! @return 0 => OK, != 0 => fail
//!
USBDM_API void prefix(write_reg)(unsigned char regIndex, unsigned long int value) {
	usb_data[0] = 6;	 /* send 6 bytes */
	usb_data[1] = CMD_CFVx_WRITE_REG;
	usb_data[2] = regIndex&0xff;
	usb_data[3] = (value>>24)&0xff; // Value is big-endian
	usb_data[4] = (value>>16)&0xff;
	usb_data[5] = (value>>8)&0xff;
	usb_data[6] = (value)&0xff;
	print("write_reg(): %s(0x%04X) <= 0x%08lX\r\n",
	                    getCFRegName(regIndex), regIndex, value);
	bdm_usb_send_ep0(usb_data);
}

/* reads byte from the specified address */
/* returns 0 on success and non-zero on failure */
USBDM_API unsigned char prefix(read_mem8)(unsigned long int address, unsigned char * result) {
	usb_data[0] = 2;	 /* get 2 bytes */
	usb_data[1] = CMD_CFVx_READ_MEM8;
	usb_data[2] = (address>>24)&0xff;  // Address is big-endian
	usb_data[3] = (address>>16)&0xff;
	usb_data[4] = (address>>8)&0xff;
	usb_data[5] = (address)&0xff;
	bdm_usb_recv_ep0(usb_data);
	*result = usb_data[1];

	print("read_mem8(): Read byte from address 0x%08lX, result: 0x%02X (0x%02X)\r\n",address,*result,usb_data[0]);
	return(usb_data[0] != CMD_CFVx_READ_MEM8);
}

/* reads word from the specified address */
/* returns 0 on success and non-zero on failure */
USBDM_API unsigned char prefix(read_mem16)(unsigned long int address, unsigned int * result) {
	usb_data[0] = 3;	 /* get 3 bytes */
	usb_data[1] = CMD_CFVx_READ_MEM16;
	usb_data[2] = (address>>24)&0xff;  // Address is big-endian
	usb_data[3] = (address>>16)&0xff;
	usb_data[4] = (address>>8)&0xff;
	usb_data[5] = (address)&0xff;
	bdm_usb_recv_ep0(usb_data);

	*result = ((unsigned int)usb_data[1]<<8)+usb_data[2];
	print("read_mem16(): Read word from address 0x%08lX, result: 0x%04X (0x%02X)\r\n",address,*result,usb_data[0]);
	return(usb_data[0] != CMD_CFVx_READ_MEM16);
}

/* reads long word from the specified address */
/* returns 0 on success and non-zero on failure */
USBDM_API unsigned char prefix(read_mem32)(unsigned long int address, unsigned long int * result) {
	usb_data[0] = 5;	 /* get 5 bytes */
	usb_data[1] = CMD_CFVx_READ_MEM32;
	usb_data[2] = (address>>24)&0xff;   // Address is big-endian
	usb_data[3] = (address>>16)&0xff;
	usb_data[4] = (address>>8)&0xff;
	usb_data[5] = (address)&0xff;
	bdm_usb_recv_ep0(usb_data);
	*result = (((unsigned long int)usb_data[1])<<24)+(usb_data[2]<<16)+(usb_data[3]<<8)+usb_data[4];

	print("read_mem32(): Read long word from address 0x%08lX, result: 0x%08lX (0x%02X)\r\n",address,*result,usb_data[0]);
	return(usb_data[0] != CMD_CFVx_READ_MEM32);
}

/* writes byte at the specified address */
USBDM_API void prefix(write_mem8)(unsigned long int address, unsigned char value) {
	usb_data[0] = 6;	 /* send 6 bytes */
	usb_data[1] = CMD_CFVx_WRITE_MEM8;
	usb_data[2] = (address>>24)&0xff;   // Address is big-endian
	usb_data[3] = (address>>16)&0xff;
	usb_data[4] = (address>>8)&0xff;
	usb_data[5] = (address)&0xff;
	usb_data[6] = (value)&0xff;
	print("write_mem8(): Write byte 0x%02X to address 0x%08lX\r\n",value,address);
	bdm_usb_send_ep0(usb_data);
}

/* writes word at the specified address */
USBDM_API void prefix(write_mem16)(unsigned long int address, unsigned int value) {
	usb_data[0] = 7;	 /* send 7 bytes */
	usb_data[1] = CMD_CFVx_WRITE_MEM16;
	usb_data[2] = (address>>24)&0xff;   // Address is big-endian
	usb_data[3] = (address>>16)&0xff;
	usb_data[4] = (address>>8)&0xff;
	usb_data[5] = (address)&0xff;
	usb_data[6] = (value>>8)&0xff;   // Value is big-endian
	usb_data[7] = (value)&0xff;
	print("write_mem16(): Write word 0x%04X to address 0x%08lX\r\n",value,address);
	bdm_usb_send_ep0(usb_data);
}

/* writes long word at the specified address */
USBDM_API void prefix(write_mem32)(unsigned long int address, unsigned long int value) {
	usb_data[0] = 9;	 /* send 9 bytes */
	usb_data[1] = CMD_CFVx_WRITE_MEM32;
	usb_data[2] = (address>>24)&0xff;    // Address is big-endian
	usb_data[3] = (address>>16)&0xff;
	usb_data[4] = (address>>8)&0xff;
	usb_data[5] = (address)&0xff;
	usb_data[6] = (value>>24)&0xff;   // Value is big-endian
	usb_data[7] = (value>>16)&0xff;
	usb_data[8] = (value>>8)&0xff;
	usb_data[9] = (value)&0xff;
	print("write_mem32(): Write long word 0x%08lX to address 0x%08lX\r\n",value,address);
	bdm_usb_send_ep0(usb_data);
}

//!  (CFv2, CFv3 & CFv4) 8-bit Memory block read
//!
//! @param address   32-bit address
//! @param byteCount block size (in bytes)
//! @param buffer    Pointer to buffer for data
//! @note Uses byte accesses
//!
//! @return 0 => Success,\n !=0 => Fail
//!
USBDM_API unsigned char prefix(read_block8)( unsigned long int address,
                                             unsigned long int byteCount,
                                             unsigned char *buffer) {
   const unsigned maxBlockSize = (MAX_DATA_SIZE-9);
	unsigned int   blockSize;

	print("read_block8(): Read 0x%08lX byte(s) from address 0x%08lX:\r\n",
	      byteCount, address );

	while (byteCount > 0) {
      blockSize = byteCount;
      if (blockSize > maxBlockSize)
         blockSize = maxBlockSize;
      usb_data[0] = blockSize+1;  // Error code + data bytes
		usb_data[1] = CMD_CFVx_READ_MEMBLOCK8;
		usb_data[2] = (address>>24)&0xff;   // Address is big-endian
		usb_data[3] = (address>>16)&0xff;
		usb_data[4] = (address>>8)&0xff;
		usb_data[5] = (address)&0xff;
		bdm_usb_recv_ep0(usb_data);
		print("read_block8(): Block read, size 0x%02X (0x%02X):\r\n",
		      blockSize, usb_data[0]);
		if (usb_data[0] != CMD_CFVx_READ_MEMBLOCK8) {
	      print("read_block8(): Error!\r\n");
         return(1);
		}
      memcpy(buffer, usb_data+1, blockSize);
		//printDump(buffer, blockSize);
		byteCount -= blockSize;
      address   += blockSize;
      buffer    += blockSize;
	}
	return(0);
}

//!  (CFv2, CFv3 & CFv4) 16-bit Memory block read
//!
//! @param address   32-bit address
//! @param byteCount block size (in bytes)
//! @param buffer    Pointer to buffer for data
//! @note Uses word accesses. If the address is odd then a byte read is performed to make it even before performing the word reads.
//!                           If there is an odd byte at the end of the block it is read using a byte read.
//!
//! @return 0 => Success,\n !=0 => Fail
//!
USBDM_API unsigned char prefix(read_block16)( unsigned long int address,
                                              unsigned long int byteCount,
                                              unsigned char *buffer) {
	const unsigned maxBlockSize = ((MAX_DATA_SIZE-9)&0xfffe);
   unsigned int   blockSize;

	print( "read_block16(): Read 0x%08lX byte(s) from address 0x%08lX:\r\n",
	       byteCount, address);

	if (address&0x01) { // Odd address - read 1 byte to make even
		print("read_block16(): Address is odd, performing 1 byte read\r\n");
		if (prefix(read_mem8)(address, buffer))
		   return(1);
		byteCount--;
		address++;
		buffer++;
	}
	while (byteCount >= 2) { // Read words
      blockSize = byteCount;
      if (blockSize > maxBlockSize)
         blockSize = maxBlockSize;
      usb_data[0] = blockSize+1;  // Error code + data bytes
		usb_data[1] = CMD_CFVx_READ_MEMBLOCK16;
		usb_data[2] = (address>>24)&0xff;   // Address is big-endian
		usb_data[3] = (address>>16)&0xff;
		usb_data[4] = (address>>8)&0xff;
		usb_data[5] = (address)&0xff;
		bdm_usb_recv_ep0(usb_data);
		print( "read_block16(): Block read, size 0x%02X (0x%02X):\r\n",
		       blockSize, usb_data[0]);
		//printDump(usb_data+1, blockSize);
      memcpy(buffer, usb_data+1, blockSize);
		byteCount -= blockSize;
		address   += blockSize;
      buffer    += blockSize;
		if (usb_data[0] != CMD_CFVx_READ_MEMBLOCK16)
		   return(1);
	}
	if (byteCount) {  // Trailing odd byte
      print("read_block16(): Performing trailing 1 byte read\r\n");
      return (prefix(read_mem8)(address, buffer));
	}
	return(0);
}

//!  (CFv2, CFv3 & CFv4) 32-bit Memory block read
//!
//! @param address   32-bit address
//! @param byteCount block size (in bytes)
//! @param buffer    Pointer to buffer for data
//! @note Uses long word accesses. If the address is not aligned then a series of byte/word reads are performed to align it.
//!
//! @return 0 => Success,\n !=0 => Fail
//!
USBDM_API unsigned char prefix(read_block32)( unsigned long int address,
                                              unsigned long int byteCount,
                                              unsigned char *buffer) {
   const unsigned maxBlockSize = ((MAX_DATA_SIZE-9)&0xfffC);
   unsigned int   blockSize;
   unsigned int originalByteCount = byteCount;
   unsigned char *originalBuffer = buffer;
   U8 oddBytes = (-address)&0x03; // Odd bytes at start of block?

   print( "read_block32(): Read 0x%08lX byte(s) from address 0x%08lX:\r\n",
	       byteCount, address);

	if (oddBytes) { // Not long word aligned - read odd bytes
		print("read_block32(): Address is not long aligned, reading odd bytes\r\n");
		if (prefix(read_block16)(address, oddBytes, buffer))
		   return(1);
		byteCount -= oddBytes;
		address   += oddBytes;
		buffer    += oddBytes;
	}
   while (byteCount >= 4) { // Read long words
      blockSize = byteCount;
      if (blockSize > maxBlockSize)
         blockSize = maxBlockSize;
      usb_data[0] = blockSize+1;  // Error code + data bytes
      usb_data[1] = CMD_CFVx_READ_MEMBLOCK32;
      usb_data[2] = (address>>24)&0xff;   // Address is big-endian
      usb_data[3] = (address>>16)&0xff;
      usb_data[4] = (address>>8)&0xff;
      usb_data[5] = (address)&0xff;
      bdm_usb_recv_ep0(usb_data);
      print( "read_block32(): Block read, size 0x%02X (0x%02X):\r\n",
             blockSize, usb_data[0]);
      //printDump(usb_data+1, blockSize);
      memcpy(buffer, usb_data+1, blockSize);
      byteCount -= blockSize;
      address   += blockSize;
      buffer    += blockSize;
      if (usb_data[0] != CMD_CFVx_READ_MEMBLOCK32)
         return(1);
   }
   if (byteCount > 0) { // Read trailing odd bytes
      print("read_block32(): Trailing block read, size 0x%02X (0x%02X):\r\n",
            byteCount, usb_data[0]);
      if (prefix(read_block16)( address, byteCount, buffer) != 0)
         return 1;
   }
   printDump(originalBuffer, originalByteCount);

   return(0);
}

//!  (CFv2, CFv3 & CFv4) 8-bit Memory block write
//!
//! @param address   32-bit address
//! @param byteCount block size (in bytes)
//! @param buffer    Pointer to buffer for data
//! @note Uses byte accesses
//!
//! @return 0 => Success,\n !=0 => Fail
//!
USBDM_API unsigned char prefix(write_block8)( unsigned long int _address,
                                              unsigned long int byteCount,
                                              unsigned char *buffer) {
   const unsigned maxBlockSize = ((MAX_DATA_SIZE-14)&0xfffe);
   unsigned int   blockSize;
   U32u address = {_address};

   print("write_block8(): Write 0x%08lX byte(s) to address 0x%08lX:\r\n",
         byteCount, address);

   while (byteCount > 0) {
      blockSize = byteCount;
      if (blockSize > maxBlockSize)
         blockSize = maxBlockSize;
      usb_data[0] = blockSize+5;    // cmd + 4 bytes address + data bytes
      usb_data[1] = CMD_CFVx_WRITE_MEMBLOCK8;
      usb_data[2] = address.bytes[3];   // Address is big-endian
      usb_data[3] = address.bytes[2];
      usb_data[4] = address.bytes[1];
      usb_data[5] = address.bytes[0];
      memcpy(usb_data+6, buffer, blockSize);
      print("write_block8(): Block write, size 0x%02X:\r\n", blockSize);
      //printDump(usb_data+6, blockSize);
      bdm_usb_send_ep0(usb_data);
#ifdef WRITE_BLOCK_CHECK
      if (prefix(get_last_sts)())
         return(1);
#endif
      byteCount        -= blockSize;
      address.longword += blockSize;
      buffer           += blockSize;
   }
	return(0);
}

//!  (CFv2, CFv3 & CFv4) 16-bit Memory block write
//!
//! @param address   32-bit address
//! @param byteCount block size (in bytes)
//! @param buffer    Pointer to buffer for data
//! @note Uses word accesses. If the address is odd then a byte write is performed to make it even before performing the word reads.
//!                           If there is an odd byte at the end of the block it is written using a byte write.
//!
//! @return 0 => Success,\n !=0 => Fail
//!
USBDM_API unsigned char prefix(write_block16)( unsigned long int address,
                                               unsigned long int byteCount,
                                               unsigned char *buffer) {
   const unsigned maxBlockSize = ((MAX_DATA_SIZE-14)&0xfffe);
   unsigned int   blockSize;

	print("write_block16(): Write 0x%08lX byte(s) from address 0x%08lX:\r\n",
	      byteCount,address);

	if (address&0x01) {
		print("write_block16(): Address is odd, performing 1 byte write\r\n");
		if (prefix(write_block8)(address, 1, buffer) != 0)
			return(1);
		byteCount--;
		address++;
		buffer++;
	}
   while (byteCount >= 2) { // Write words
      blockSize = byteCount;
      if (blockSize > maxBlockSize)
         blockSize = maxBlockSize;
      usb_data[0] = blockSize+5;    // cmd + 4 bytes address + data bytes
      usb_data[1] = CMD_CFVx_WRITE_MEMBLOCK16;
      usb_data[2] = (address>>24)&0xff;   // Address is big-endian
      usb_data[3] = (address>>16)&0xff;
      usb_data[4] = (address>>8)&0xff;
      usb_data[5] = (address)&0xff;
      memcpy(usb_data+6, buffer, blockSize);
      print("write_block16(): Block write, size 0x%02X:\r\n", blockSize);
      //printDump(usb_data+6, blockSize);
      bdm_usb_send_ep0(usb_data);
#ifdef WRITE_BLOCK_CHECK
      if (prefix(get_last_sts)())
         return(1);
#endif
      byteCount -= blockSize;
      address   += blockSize;
      buffer    += blockSize;
   }
	if (byteCount>0) {	/* leftover byte */
		print("write_block16(): Misaligned byte at the end of the block, performing 1 byte write\r\n");
		return (prefix(write_block8)(address, 1, buffer));
	}
	return(0);
}

//!  (CFv2, CFv3 & CFv4) 32-bit Memory block write
//!
//! @param address   32-bit address
//! @param byteCount block size (in bytes)
//! @param buffer    Pointer to buffer for data
//! @note Uses long word accesses. If the address is not aligned then a series of byte/word writes are performed to align it.
//!                                If there are odd bytes at the end of the block they is written using byte/word writes.
//!
//! @return 0 => Success,\n !=0 => Fail
//!
USBDM_API unsigned char prefix(write_block32)( unsigned long int address,
                                               unsigned long int byteCount,
                                               unsigned char *buffer) {
   const unsigned maxBlockSize = ((MAX_DATA_SIZE-14)&0xFFFC);
   unsigned int   blockSize;
   U8 oddBytes = (-address)&0x03; // Odd bytes at start of block?

	print("write_block32(): Write 0x%08lX byte(s) at address 0x%08lX:\r\n",byteCount,address);

	//printDump(buffer, byteCount);
	if (oddBytes) {
		print("write_block32(): Address is misaligned, performing word/byte writes\r\n");
		if(prefix(write_block16)(address, oddBytes, buffer) != 0)
         return(1);
		byteCount -= oddBytes;
		buffer    += oddBytes;
      address   += oddBytes;
	}
   while (byteCount >= 4) { // Long word writes
      blockSize = byteCount;
      if (blockSize > maxBlockSize)
         blockSize = maxBlockSize;
      usb_data[0] = blockSize+5;    // cmd + 4 bytes address + data bytes
      usb_data[1] = CMD_CFVx_WRITE_MEMBLOCK32;
      usb_data[2] = (address>>24)&0xff;  // Address is big-endian
      usb_data[3] = (address>>16)&0xff;
      usb_data[4] = (address>>8)&0xff;
      usb_data[5] = (address)&0xff;
      memcpy(usb_data+6, buffer, blockSize);
      print("write_block32(): Block write, size 0x%02X:\r\n", blockSize);
      //printDump(usb_data+6, blockSize);
      bdm_usb_send_ep0(usb_data);
#ifdef WRITE_BLOCK_CHECK
      if (prefix(get_last_sts)())
         return(1);
#endif
      byteCount -= blockSize;
      address   += blockSize;
      buffer    += blockSize;
   }
	if (byteCount>0) {	/* leftover word/bytes */
		print("write_block32(): Partial longword at the end of the block\r\n");
		return prefix(write_block16)(address, byteCount, buffer);
	}
	return(0);
}

//! JTAG - move the TAP from \b RUN-TEST/IDLE to \b SHIFT-DR or \b SHIFT-IR state
//!
//! @param mode \n - ==0 : go to \b SHIFT-DR, \n - !=0 : go to \b SHIFT-IR
//! @return 0 on success and non-zero on failure
//! @note (requires the tap to initially be in \b RUN-TEST/IDLE)
//!
USBDM_API unsigned char prefix(jtag_sel_shift)(unsigned char mode) {
	usb_data[0]=1;	 /* get 1 byte */
	usb_data[1]=CMD_JTAG_GOTOSHIFT;
	usb_data[2]=mode;
	bdm_usb_recv_ep0(usb_data);

	print("TBLCF_JTAG_SEL_SHIFT: SHIFT-%cR\r\n",mode?'I':'D');
	return(usb_data[0] != CMD_JTAG_GOTOSHIFT);
}

//! JTAG - Takes the TAP to \b TEST-LOGIC-RESET state
//!
//! @note use set_target_type() with \ref T_JTAG target to take TAP back to \b RUN-TEST/IDLE
//! @return 0 on success and non-zero on failure
//!
unsigned char USBDM_API prefix(jtag_sel_reset)(void) {
	usb_data[0]=1;	 /* get 1 byte */
	usb_data[1]=CMD_JTAG_GOTORESET;
	bdm_usb_recv_ep0(usb_data);

	print("TBLCF_JTAG_SEL_RESET\r\n");
	return(usb_data[0] != CMD_JTAG_GOTORESET);
}

//! JTAG - write data to JTAG shift register
//!
//!  @param exit \n - ==0 : stay in \b SHIFT-xx, \n - !=0 : go to \b RUN-TEST/IDLE, \n
//!  @param bit_count 8-bit count of \b bits to shift in
//!  @param buffer Pointer to data buffer.  The data is shifted in LSB (last byte) first,
//!                unused bits (if any) are in the MSB (first) byte;
//!  @note Requires the tap to be in \b SHIFT-DR or \b SHIFT-IR state.
//!
USBDM_API void prefix(jtag_write)(unsigned char bit_count,
                                  unsigned char exit,
                                  unsigned char *buffer) {
	int i;
	usb_data[0]=(bit_count>>3)+((bit_count&0x07)!=0)+3;	 /* send all data bits & 3 more bytes */
	usb_data[1]=CMD_JTAG_WRITE;
	usb_data[2]=exit;
	usb_data[3]=bit_count;
	for (i=0;i<((bit_count>>3)+1);i++)
	   usb_data[4+i]=*(buffer+i);	/* copy data */
	bdm_usb_send_ep0(usb_data);

	print("TBLCF_JTAG_WRITE: %d bits (%sexit to RUN-TEST/IDLE)\r\n",bit_count,exit?"":"NO ");
	printDump(usb_data+4,(bit_count>>3)+((bit_count&0x07)!=0));
}

//! JTAG - read data from JTAG shift register
//! @param exit - ==0 : stay in \b SHIFT-xx, \n - !=0 : go to \b RUN-TEST/IDLE, \n
//! @param bit_count Count of \b bits to shift out
//! @param buffer Pointer to buffer for data read out of the device (first bit in LSB of the last byte in the buffer)
//! @note Requires the tap to be in \b SHIFT-DR or \b SHIFT-IR state.
//! @return 0 on success and non-zero on failure
//!
USBDM_API unsigned char prefix(jtag_read)( unsigned char bit_count,
                                           unsigned char exit,
                                           unsigned char *buffer) {
	int i;
	usb_data[0]=(bit_count>>3)+((bit_count&0x07)!=0)+1;	 /* get all data bits & the status byte */
	usb_data[1]=CMD_JTAG_READ;
	usb_data[2]=exit;
	usb_data[3]=bit_count;
	bdm_usb_recv_ep0(usb_data);
	for (i=0;i<((bit_count>>3)+((bit_count&0x07)!=0));i++)
	   *(buffer+i)=usb_data[1+i];	/* copy data */
	print("TBLCF_JTAG_READ: %d bits (%sexit to RUN-TEST/IDLE)\r\n",bit_count,exit?"":"NO ");
	printDump(usb_data+1,(bit_count>>3)+((bit_count&0x07)!=0));
	return( usb_data[0] != CMD_JTAG_READ);
}
#endif // TBLCF

//=============================================================================================
// TBDML & OSBDM common dll entry points
//
#if defined(TBDML) || defined(OSBDM)
//! Determine crystal frequency of the target in MHz
//!
//! @return Float representing the Crystal Frequency of target
//!
USBDM_API float prefix(get_speed)(void) {
   usb_data[0] = 3;  // receive up to 3 bytes
	usb_data[1] = CMD_READ_SPEED;
	if (bdm_usb_recv_ep0(usb_data) != BDM_RC_OK)
	   return 8E6; // default to safe 8MHz on error
	return (60.0 * 2 * 128 / (*(U16 *)(usb_data+1)));
}

//! Sets the communication speed.
//!
//! @param crystal_frequency crystal frequency in MHz.
//! This is used to determine the BDM communication speed (assumed
//! to be half the crystal frequency given).
//!
//! @return 0 => Success,\n !=0 => Fail
//!
USBDM_API unsigned char prefix(set_speed)(float crystal_frequency) {
	/* the command parameter is the expected length
	 *  of SYNC pulse expressed in 60MHz ticks */
   usb_data[0] = 1;  // receive up to 1 byte
	usb_data[1] = CMD_SET_SPEED1;
	*(U16 *)(usb_data+2) = 60.0 * 2 * 128 / crystal_frequency;

	bdm_usb_recv_ep0(usb_data);
	return (usb_data[0] != CMD_SET_SPEED1);
}

//! (HCS12, HC08 & RS08) Reads one byte from memory
//!
//! @param address 16-bit memory address
//!
//! @return 8-bit value
//!
USBDM_API unsigned char prefix(read_byte)(unsigned int address) {
   usb_data[0]          = 2;  // receive up to 2 bytes
	usb_data[1]          = CMD_READ_8;
	*(U16 *)(usb_data+2) = address;
	bdm_usb_recv_ep0(usb_data);
	return (usb_data[1]);
}

//! (HCS12, HC08 & RS08) Writes one byte to memory
//!
//! For HCS12 & HC08 this writes to RAM.
//! For RS08 this may write to RAM or Flash if preceded by a call to mem_dlstart().
//!
//! @param address 16-bit memory address
//! @param data    8-bit data value
//!
USBDM_API void prefix(write_byte)( unsigned int  address,
                                   unsigned char data) {
   usb_data[0]          = 1;  // receive up to 1 byte
	usb_data[1]          = CMD_WRITE_8;
	*(U16 *)(usb_data+2) = address;
	usb_data[4]          = data;
	bdm_usb_recv_ep0(usb_data);
}

//! (HCS12, HC08 & RS08) Reads block from memory
//!
//! @param addr    16-bit memory address
//! @param count   8-bit byte count
//! @param data    Pointer to buffer to contain data
//!
USBDM_API void prefix(read_block)( unsigned int addr,
                                   unsigned int count,
                                   unsigned char *data) {

   unsigned int blockSize;
   // Make multiple of 2, Allow for status[0]+data[1...]
   const unsigned int MaxDataSize = (MAX_PACKET_SIZE-1) & ~1;

   while (count > 0) {
      blockSize = count;
      if (blockSize > MaxDataSize)
         blockSize = MaxDataSize;

      /* if address is odd and we are going to read more than one
       *  block make the first packet 1 byte short to align the address */
      if ((count>MaxDataSize) && (addr&0x1))
         blockSize--;

      usb_data[1]             = CMD_READ_BLOCK1;
      *((U16 *)(usb_data+2))  = addr;
      usb_data[4]             = blockSize;
      #ifdef FAST_BLOCK
         usb_data[0] = 4;
         tbdml_usb_send_ep2(usb_data);
         usb_data[0] = blockSize+1;
         tbdml_usb_recv_ep2(usb_data);
      #else
         usb_data[0] = blockSize+1;
         bdm_usb_recv_ep0(usb_data);
      #endif
      memcpy(data, usb_data+1, blockSize);

      data  += blockSize;   // update location in buffer
      addr  += blockSize;   // update memory address
      count -= blockSize;   // update count
   }
}

//! (HCS12, HC08 & RS08) Writes a block of data to memory
//!
//! For HCS12 & HC08 this writes to RAM.
//! For RS08 this may write to RAM or Flash if preceded by prefix(mem_dlstart().
//!
//! @param addr    16-bit memory address
//! @param count   8-bit byte count
//! @param data    Pointer to buffer containing data
//!
USBDM_API void prefix(write_block)( unsigned int addr,
                                    unsigned int count,
                                    unsigned char *data) {
unsigned retry;

   unsigned int blockSize;
   // Make multiple of 2, allow for status[0]+setup[1..4] + data[5...]
   const unsigned int MaxDataSize = (MAX_PACKET_SIZE-5)&~1;

   while (count > 0) {
      blockSize = count;
      if (blockSize > MaxDataSize)
         blockSize = MaxDataSize;

      /* if address is odd and we are going to read more than one
       *  block make the first packet 1 byte short to align the address */
      if ((count>MaxDataSize)&&(addr&0x1))
         blockSize--;

      usb_data[0]             = blockSize+4;
      usb_data[1]             = CMD_WRITE_BLOCK1;
      *((U16 *)(usb_data+2))  = addr;
      usb_data[4]             = blockSize;
      memcpy(usb_data+5, data, blockSize);

      #ifdef FAST_BLOCK
         tbdml_usb_send_ep2(usb_data); /* Tx data */
         usb_data[0]=1;                /* read status */
         tbdml_usb_recv_ep2(usb_data);
      #else
         // ToDo - error check
         bdm_usb_send_ep0(usb_data);
         // do handshake
         retry = 10;
         do {
            usb_data[0]=1;  /* receive 1 byte */
            usb_data[1]=CMD_GET_LAST_STATUS;
            if (bdm_usb_recv_ep0(usb_data) == 0)
                break;
            Sleep(50 /* ms */); // So we don't tie up the USB
         } while (retry-- > 0);
      #endif

      data  += blockSize;   // update location in buffer
      addr  += blockSize;   // update address
      count -= blockSize;   // update count
   }
}

//! (HCS12) Reads one word from memory (address must be aligned)
//!
//! @param address 16-bit memory address
//!
//! @return 16-bit data value
//!
USBDM_API unsigned int prefix(read_word)(unsigned int address) {
   usb_data[0] = 3;  // receive up to 3 bytes
	usb_data[1] = CMD_READ_16;
	*((U16 *)(usb_data+2)) = address;
	bdm_usb_recv_ep0(usb_data);
	return (*(U16 *)(usb_data+1));
}

//! (HCS12) Writes one word to memory (address must be aligned)
//!
//! @param address 16-bit memory address
//! @param data 16-bit data value
//!
USBDM_API void prefix(write_word)(unsigned int address, unsigned int data) {
   usb_data[0] = 1;  // receive up to 1 byte
	usb_data[1] = CMD_WRITE_16;
	*(U16 *)(usb_data+2) = address;
	*(U16 *)(usb_data+4) = data;
	bdm_usb_recv_ep0(usb_data);
}

#define HC12_BDMSTS     0xFF01   //!< Address of HC12 BDM Status register

//! (HCS12) Reads one byte from the BDM address space
//!
//! @param address 16-bit memory address
//! @return 8-bit data value
//!
USBDM_API unsigned char prefix(read_bd)(unsigned int address) {
   unsigned char BDMStatus;
   usb_data[0] = 2;  // receive up to 2 bytes
	usb_data[1] = CMD_READ_BD;
	*((U16 *)(usb_data+2)) = address;
	bdm_usb_recv_ep0(usb_data);
	BDMStatus = usb_data[1];
   if (address == HC12_BDMSTS) {
      print("Status = %s, %s, %s\r\n", (BDMStatus&0x80)?"ENBDM":"-",
                                       (BDMStatus&0x40)?"BDACT":"-",
                                       (BDMStatus&0x04)?"CLKSW":"-" );
	}
	return (usb_data[1]);
}

//! (HCS12) Writes one byte to the BDM address space
//!
//! @param address 16-bit memory address
//! @param data    8-bit data value
//!
//! @return 0 => Success,\n !=0 => Fail
//!
USBDM_API unsigned char prefix(write_bd)(unsigned int address, unsigned char data) {
   usb_data[0] = 2;  // receive up to 2 bytes
	usb_data[1] = CMD_WRITE_BD;
	*((U16 *)(usb_data+2)) = address;
	usb_data[4] = data;
	bdm_usb_recv_ep0(usb_data);
	return (usb_data[0]!=CMD_WRITE_BD);
}
#endif // defined(TBDML) || defined(OSBDM)

//=============================================================================================
//  OSBDM dll entry points
//
#ifdef OSBDM
//! (RS08) Sets target derivative type
//!
//! @param derivative type of RFS08 derivative, see \ref rs08_derivative_type_e
//!
//! @return 0 => Success,\n !=0 => Fail
//!
USBDM_API unsigned char prefix(set_derivative_type)(unsigned char derivative) {

   // Dummy this routine until Codewarrior supports RS08 properly
   return 0;

#if 0
   usb_data[0] = 1;  // receive up to 1 byte
   usb_data[1] = CMD_SET_DERIVATIVE;
   usb_data[2] = derivative;
   bdm_usb_recv_ep0(usb_data);
   return (usb_data[0]!=CMD_SET_DERIVATIVE);
#endif
}

//===========================================================
//  RS08 Flash memory handling
//
// ToDo - review Flash handling
//

//! (RS08) Prepare for RS08 Flash programming (Vpp on & Mass erase Flash)
//!
//! prefix(write_block() and prefix(write_byte() may now be
//! be used to program blocks of data or individual bytes to the RS08 Flash.
//!
//! \warning Needs review
//!
USBDM_API void prefix(mem_dlstart)(void){
   usb_data[0] = 1;  // receive up to 2 bytes
   usb_data[1] = CMD_FLASH_DLSTART;
   bdm_usb_recv_ep0(usb_data);
}

//! (RS08) Check if ready for next block to program to Flash
//!
//! \warning I think this is broken since the BDM USB interface
//!          can't respond while busy programming Flash.
//!
USBDM_API unsigned char prefix(target_readyfor_datald)(void){
   usb_data[0] = 2;  // receive up to 2 bytes
   usb_data[1] = TCMD_NEXT_DLDATA;
   if (bdm_usb_recv_ep0(usb_data) != 0)
      return 1;

   return ((usb_data[0] == TCMD_NEXT_DLDATA) &&
          (usb_data[1] != BDM_RC_FLASH_PROGRAMING_BUSY));
}

//! (RS08) Tidy up after Flash programming \n
//! Restore/program clock trim value, Vpp off
//!
//! \warning Needs review
//!
USBDM_API void prefix(mem_dlend)(void){
   usb_data[0] = 2;  // receive up to 2 bytes
   usb_data[1] = CMD_FLASH_DLEND;
   bdm_usb_recv_ep0(usb_data);
}

//=========================================================
// Memory access
//=========================================================

//!  (CFv1) Memory read
//!
//! @param addr      24-bit address
//! @param size      block size (in bytes)
//! @param memSpace  Memory transfer size (1,2 or 4)
//! @param data      Pointer to buffer for data
//!
//! @return 0 => Success,\n !=0 => Fail
//!
BOOL USBDM_API prefix(read_mem)( unsigned int addr,     unsigned int size,
                                 unsigned int memSpace, unsigned char *data) {
   unsigned int blockSize;
   unsigned int count = size;
   // Make multiple of 4, Allow for status[0]+data[1...]
   const unsigned int CF_MAX_DATA_SIZE = (MAX_PACKET_SIZE-1)&~3;
   BOOL unaligned;

   print("prefix(read_mem(addr=0x%X,size=%d,space=%d)\r\n",
                                  addr, size, memSpace);
   // Check address alignment
   switch (memSpace) {
      case 1: unaligned = 0;               // No alignment requirement
         break;
      case 2: unaligned = ((addr&1) != 0) || ((size&1) != 0); // Multiple of 2
         break;
      case 4: unaligned = ((addr&3) != 0) || ((size&3) != 0); // Multiple of 4
         break;
      default: unaligned = 1;              // ERROR!
         return FALSE;
   }
   if (unaligned) {
      print("prefix(read_mem - alignment error\r\n");
      return FALSE;
   }

   while (count>0) {
      blockSize = count;
      if (blockSize > CF_MAX_DATA_SIZE)
         blockSize = CF_MAX_DATA_SIZE;
      usb_data[0] = (U8)(blockSize+1);    // max # of bytes in USB response
      usb_data[1] = CMD_CF_READ_MEM;      //    status[0] + data[1...]
      usb_data[2] = (U8)addr;
      usb_data[3] = (U8)(addr>>8);
      usb_data[4] = (U8)(addr>>16);
      usb_data[5] = (U8)memSpace;
      // USB Handler in BDM calculates data blockSize from transfer size
      // and places it here
      //usb_data[6] = (U8) blockSize;
      if (bdm_usb_recv_ep0(usb_data) != 0)
         return FALSE;
      memcpy(data, usb_data+1, blockSize);

      data  += blockSize;   // update destination in buffer
      addr  += blockSize;   // update address
      count -= blockSize;   // update count
   }
   print("========read_mem========\r\n");
   printDump(data-size, size);

   return TRUE; //ToDo - Check this!!!
}

//!  (CFv1) Memory write
//!
//! @param addr      24-bit address
//! @param count     8-bit block size (in bytes)
//! @param memSpace  Memory transfer size (1,2 or 4)
//! @param data      Pointer to buffer containing data
//!
//! @return 0 => Success,\n !=0 => Fail
//!
BOOL USBDM_API prefix(write_mem)( unsigned int addr,     unsigned int count,
                                  unsigned int memSpace, unsigned char *data) {
   char tt[10];
   unsigned int blockSize;
   // Make multiple of 4, allow for status[0]+setup[1..6] + data[7...]
   const unsigned int MaxDataSize = (MAX_PACKET_SIZE-7)&~3;
   BOOL unaligned;

   print("prefix(write_mem(addr=0x%X,count=%d,space=%d)\r\n",
                                   addr, count, memSpace);
   // Check address alignment
   switch (memSpace) {
      case 1: unaligned = 0;               // No alignment requirement
         break;
      case 2: unaligned = ((addr&1) != 0) || ((count&1) != 0); // Multiple of 2
         break;
      case 4: unaligned = ((addr&3) != 0) || ((count&3) != 0); // Multiple of 4
         break;
      default: unaligned = 1;              // ERROR!
         return FALSE;
   }
   if (unaligned) {
      print("prefix(write_mem - alignment error\r\n");
      return FALSE;
   }

   printDump(data, count);

   while (count>0) {
      blockSize = count;
      if (blockSize > MaxDataSize)
         blockSize = MaxDataSize;
      usb_data[0] = (U8)(blockSize+6);  // # of bytes to send
      usb_data[1] = CMD_CF_WRITE_MEM;   //   setup[1..6] + data[7....]
      usb_data[2] = (U8)addr;
      usb_data[3] = (U8)(addr>>8);
      usb_data[4] = (U8)(addr>>16);
      usb_data[5] = (U8)memSpace;
      usb_data[6] = (U8)blockSize;
      memcpy(usb_data+7, data, blockSize);
      if (bdm_usb_send_ep0(usb_data))
         return FALSE;

      // do handshake
      tt[0]=1;  /* receive 1 byte */
      tt[1]=CMD_GET_LAST_STATUS;
      bdm_usb_recv_ep0(tt);

      data  += blockSize;   // update location in buffer
      addr  += blockSize;   // update memory address
      count -= blockSize;   // update count
   }
   return TRUE;
}

//==========================================================
//  Register Access
//==========================================================

//! (CFv1) Read data or address register
//!
//! @param regNo  Register #
//!
//! @return 32-bit number
//!
USBDM_API unsigned int  prefix(read_reg)(unsigned char regNo) {
   print("prefix(read_reg(%d)\r\n", regNo);
   usb_data[0] = 5;  // receive up to 5 bytes
   usb_data[1] = CMD_CF_READ_REG;
   usb_data[2] = regNo;
   bdm_usb_recv_ep0(usb_data);
   return (*(U32 *)(usb_data+1));
}

//! (CFv1) Read Control register
//!
//! @param regNo  Register #
//!
//! @return 32-bit number
//!
USBDM_API unsigned int  prefix(read_creg)(unsigned char regNo) {
   print("prefix(read_creg(%d)\r\n", regNo);
   usb_data[0] = 5;  // receive up to 5 bytes
   usb_data[1] = CMD_CF_READ_CREG;
   usb_data[2] = regNo;
   bdm_usb_recv_ep0(usb_data);
   return (*(U32 *)(usb_data+1));
}

//! (CFv1) Read Debug register
//!
//! @param regNo  Register #
//!
//! @return 32-bit number
//!
USBDM_API unsigned int  prefix(read_dreg)(unsigned char regNo) {
   usb_data[0] = 5;  // receive up to 5 bytes
   usb_data[1] = CMD_CF_READ_DREG;
   usb_data[2] = regNo;
   bdm_usb_recv_ep0(usb_data);
   print("read_dreg(): %s(0x%04X) == 0x%08lX\r\n",
         getCFDebugRegName(regNo), regNo, (*(U32 *)(usb_data+1)));
   return (*(U32 *)(usb_data+1));
}

//! (CFv1) Read XCSR.msb
//!
//! @return 8-bit number
//!
//! @note This is mapped to prefix(read_status() by the BDM firmware
//
USBDM_API unsigned int  prefix(read_xcsr_byte)(void) {
   unsigned char BDMStatus;
   char *runMode[] = {"Running","Stopped","Halted","Illegal"};
   print("prefix(read_xcsr_byte()\r\n");
   usb_data[0] = 2;  // receive up to 2 bytes
   usb_data[1] = CMD_CF_READ_XCSR;
   bdm_usb_recv_ep0(usb_data);
	BDMStatus = usb_data[1];
   print("Status = %s, %s, %s, %s, %s, %s\r\n", runMode[(BDMStatus>>6)&0x03],
                                                (BDMStatus&0x20)?"Invalid":"-",
                                                (BDMStatus&0x10)?"Illegal":"-",
                                                (BDMStatus&0x08)?"Overrun(busy)":"-",
                                                (BDMStatus&0x01)?"ENBDM":"-",
                                                (BDMStatus&0x04)?"CLKSW":"-" );
   return (usb_data[1]);
}

#if 0
//! (CFv1) Read CSR2.msb
//!
//! @return 8-bit number
//!
USBDM_API unsigned int  prefix(read_csr2_byte)(void) {
   print("prefix(read_csr2_byte()\r\n");
   usb_data[0] = 2;  // receive up to 2 bytes
   usb_data[1] = CMD_CF_READ_XCSR;
   bdm_usb_recv_ep0(usb_data);
   return (usb_data[1]);
}

//! (CFv1) Read CSR3.msb
//!
//! @return value 8-bit number
//!
USBDM_API unsigned int  prefix(read_csr3_byte)(void) {
   print("prefix(read_csr3_byte()\r\n");
   usb_data[0] = 2;  // receive up to 2 bytes
   usb_data[1] = CMD_CF_READ_XCSR;
   bdm_usb_recv_ep0(usb_data);
   return (usb_data[1]);
}
#endif

//! (CFv1) Write XCSR.msb
//!
//! @param value 8-bit number
//!
//! @note This is mapped to prefix(write_status() by the BDM firmware
//
USBDM_API void  prefix(write_xcsr_byte)(unsigned char value) {
	usb_data[0] = 1; // receive 1 byte
	usb_data[1] = CMD_CF_WRITE_XCSR;
	usb_data[2] = value;
	bdm_usb_recv_ep0(usb_data);
}

#if 0
//! (CFv1) Write CSR2.msb
//!
//! @param value 8-bit number
//!
USBDM_API void  prefix(write_csr2_byte)(unsigned char value) {
   usb_data[0] = 1; // receive 1 byte
	usb_data[1] = CMD_CF_WRITE_CSR2;
	usb_data[2] = value;
	bdm_usb_recv_ep0(usb_data);
}

//! (CFv1) Write CSR3.msb
//!
//! @param value 8-bit number
//!
USBDM_API void  prefix(write_csr3_byte)(unsigned char value) {
   usb_data[0] = 1; // receive 1 byte
	usb_data[1] = CMD_CF_WRITE_CSR3;
	usb_data[2] = value;
	bdm_usb_recv_ep0(usb_data);
}
#endif

//! (CFv1) Write Address or Data Register
//!
//! @param regNo Register #
//! @param value 32-bit value
//!
USBDM_API void prefix(write_reg)(unsigned char regNo, unsigned int value) {
   char tt[2];
   print("prefix(write_reg(%d,%X)\r\n", regNo, value);
   if (regNo>15) {// debugger tries to write to illegal registers !
      print("prefix(write_reg(%d,%X) - illegal register\r\n", regNo, value);
      return;
   }
   usb_data[0] = 6;  // send 6 bytes
   usb_data[1] = CMD_CF_WRITE_REG;
   usb_data[2] = regNo;
   *((U32 *)(usb_data+3)) = value;
   bdm_usb_send_ep0(usb_data);

   // do handshake
   tt[0] = 1;  /* receive 1 byte */
   tt[1] = CMD_GET_LAST_STATUS;
   bdm_usb_recv_ep0(tt);
}

//! (CFv1) Write Control register
//!
//! @param regNo Register #
//! @param value 32-bit value
//!
USBDM_API void prefix(write_creg)(unsigned char regNo, unsigned int value) {
   char tt[20];
   print("prefix(write_creg(%d,%X)\r\n", regNo, value);
   if (regNo>15) {// debugger tries to write to illegal registers !
      print("prefix(write_creg(%d,%X) - illegal register\r\n", regNo, value);
      return;
   }
   usb_data[0] = 6;  // send 6 bytes
   usb_data[1] = CMD_CF_WRITE_CREG;
   usb_data[2] = regNo;
   *((U32 *)(usb_data+3)) = value;
   bdm_usb_send_ep0(usb_data);

   // do handshake
   tt[0] = 1;  /* receive 1 byte */
   tt[1] = CMD_GET_LAST_STATUS;
   bdm_usb_recv_ep0(tt);
}

//! (CFv1) Write Debug register
//!
//! @param regNo Register #
//! @param value 32-bit value
//!
USBDM_API void prefix(write_dreg)(unsigned char regNo, unsigned int value) {
   char tt[20];
   print("prefix(write_dreg(%d,%X)\r\n", regNo, value);
   if (regNo>31) {// debugger tries to write to illegal registers !
      print("prefix(write_dreg(%d,%X) - illegal register\r\n", regNo, value);
      return;
   }
   usb_data[0] = 6;  // send 6 bytes
   usb_data[1] = CMD_CF_WRITE_DREG;
   usb_data[2] = regNo;
   *((U32 *)(usb_data+3)) = value;
   bdm_usb_send_ep0(usb_data);

   // do handshake
   tt[0] = 1;  /* receive 1 byte */
   tt[1] = CMD_GET_LAST_STATUS;
   bdm_usb_recv_ep0(tt);
}
#endif

#if defined(TBDML) || defined(OSBDM)
//! (HCS12, HCS08 & RS08) Reads contents of core registers
//!
//! @param  registers   pointer to structure to receive the register values
//!
//! @return 0 => Success,\n !=0 => Fail
//!
USBDM_API unsigned char prefix(read_regs)(registers_t *registers) {
   usb_data[0] = 13;  // receive up to 13 bytes
	usb_data[1] = CMD_READ_REGS;
	bdm_usb_recv_ep0(usb_data);
#ifdef TBDML // HCS12
	registers->hc12.pc  = (*((U16*)(usb_data+1)));
	registers->hc12.sp  = (*((U16*)(usb_data+3)));
	registers->hc12.ix  = (*((U16*)(usb_data+5)));
	registers->hc12.iy  = (*((U16*)(usb_data+7)));
	registers->hc12.d   = (*((U16*)(usb_data+9)));
	registers->hc12.ccr = usb_data[11];
#else // HCS08/RS08
   // Note: HCS08/RS08 structs are compatible
	registers->hcs08.pc  = *(U16 *)(usb_data+1);
	registers->hcs08.sp  = *(U16 *)(usb_data+3);
	registers->hcs08.hx  = *(U16 *)(usb_data+5);
	registers->hcs08.a   = usb_data[7];
	registers->hcs08.ccr = usb_data[8];
#endif
	return (usb_data[0]!=CMD_READ_REGS);
}

//! (HCS12, HCS08 & RS08) Writes PC (HCS12, HC08), CCR+PC (RS08)
//!
//! @param value 16-bit value
//!
USBDM_API void prefix(write_reg_pc)(unsigned int value) {
   usb_data[0] = 1;  // receive up to 1 byte
	usb_data[1] = CMD_WRITE_REG_PC;
	*(U16 *)(usb_data+2) = value;
	bdm_usb_recv_ep0(usb_data);
}

//! (HCS12, HCS08 & RS08) Writes SP (HCS12, HCS08), Shadow PC (RS08)
//!
//! @param value 16-bit value
//!
USBDM_API void prefix(write_reg_sp)(unsigned int value) {
   usb_data[0] = 1;  // receive up to 1 byte
	usb_data[1] = CMD_WRITE_REG_SP;
	*(U16 *)(usb_data+2) = value;
	bdm_usb_recv_ep0(usb_data);
}

//! (HCS12 & HCS08) Writes IX (HCS12), H:X (HCS08)
//!
//! @param value 16-bit value
//!
USBDM_API void prefix(write_reg_x)(unsigned int value) {
   usb_data[0] = 1;  // receive up to 1 byte
	usb_data[1] = CMD_WRITE_REG_X;
	*(U16 *)(usb_data+2) = value;
	bdm_usb_recv_ep0(usb_data);
}

//! (HCS12) Writes IY
//!
//! @param value 16-bit value
//!
USBDM_API void prefix(write_reg_y)(unsigned int value) {
   usb_data[0] = 1;  // receive up to 1 byte
	usb_data[1] = CMD_WRITE_REG_Y;
	*(U16 *)(usb_data+2) = value;
	bdm_usb_recv_ep0(usb_data);
}

//! (HCS12, HCS08 & RS08) Writes D=B:A (HCS12) or A (HCS08, RS08)
//!
//! @param value 16-bit / 8-bit value
//!
USBDM_API void prefix(write_reg_d)(unsigned int value) {
   usb_data[0] = 1;  // receive up to 1 byte
	usb_data[1] = CMD_WRITE_REG_D;
	*(U16 *)(usb_data+2) = value;
	bdm_usb_recv_ep0(usb_data);
}

//! (HCS12 & HCS08) Writes CCR
//!
//! @param value 8-bit value
//!
USBDM_API void prefix(write_reg_ccr)(unsigned int value) {
   usb_data[0] = 1;  // receive up to 1 byte
	usb_data[1] = CMD_WRITE_REG_CCR;
	*(U16 *)(usb_data+2) = value;
	bdm_usb_recv_ep0(usb_data);
}

//! (HCS12, HCS08 & RS08) Read target BDM status register (BDCSC, BDMSTS or XCSR as appropriate)
//!
//! @return 8-bit value
//!
//!
USBDM_API unsigned char prefix(read_status)(void){
   unsigned char BDMStatus;
   usb_data[0] = 2;  // receive up to 2 bytes
   usb_data[1] = CMD_READ_STATUS;
   bdm_usb_recv_ep0(usb_data);
	BDMStatus = usb_data[1];
   print("Status = %s, %s, %s, %s\r\n", (BDMStatus&0x80)?"ENBDM":"-",
                                        (BDMStatus&0x40)?"BDACT":"-",
                                        (BDMStatus&0x20)?"BKPTEN":"-",
                                        (BDMStatus&0x08)?"CLKSW":"-" );
   return (usb_data[1]);
}

//! (HCS12, HCS08 & RS08) Write to BDM Control Register (BDCSC, BDMSTS or XCSR as appropriate)
//!
//! @param value 8-bit value
//!
USBDM_API void prefix(write_control)(unsigned char value){
   usb_data[0] = 1;  // receive up to 1 byte
   usb_data[1] = CMD_WRITE_CONTROL;
   usb_data[2] = value;
   bdm_usb_recv_ep0(usb_data);
}

//! (HCS08 & RS08) Write to Breakpoint Register
//!
//! @param value 16-bit value
//!
USBDM_API void prefix(write_bkpt)(unsigned int value){
   usb_data[0] = 1;  // receive up to 1 byte
   usb_data[1] = CMD_WRITE_BKPT;
   *(U16 *)(usb_data+2) = (U16)value;
   bdm_usb_recv_ep0(usb_data);
}

//! (HCS08 & RS08) Read from Breakpoint Register
//!
//! @return 16-bit value
//!
USBDM_API unsigned int prefix(read_bkpt)(void){
   usb_data[0] = 3;  // receive up to 3 bytes
   usb_data[1] = CMD_READ_BKPT;
   bdm_usb_recv_ep0(usb_data);
   return (*(U16*)(usb_data+1));
}
#endif // OSBDM


//==================================================================================
// Common ICP routines
//

/*! Set BDM for ICP mode & immediately reboot
 *
 *  The BDM resets in ICP mode after this command
 *  @note Communication is lost.
 */
USBDM_API void prefix(setBoot)(void){

   usb_data[0] = 5;  // send 6 bytes
   usb_data[1] = CMD_SET_BOOT;
   usb_data[2] = 'B';
   usb_data[3] = 'O';
   usb_data[4] = 'O';
   usb_data[5] = 'T';

   (void)bdm_usb_send_ep0(usb_data); // USB transfer will be incomplete
}

//! ICP mode - reboot
//!
//! The BDM does a normal reset
//!
//! Used in ICP mode to reset to normal (BDM) mode.
//! @note Communication is lost.  \n
//!       This will do the usual flash validity checks before entering BDM mode.
//!
USBDM_API void prefix(icp_reboot)( void ) {

   bdm_usb_raw_send_ep0(ICP_REBOOT,
                        0,
                        0,
                        0,
                        NULL);
}

//! ICP mode - program BDM Flash memory
//!
//! @param addr    32-bit memory address
//! @param count   number of bytes to program
//! @param data    Pointer to buffer containing data
//!
USBDM_API unsigned int prefix(icp_program)( unsigned int  addr,
                                            unsigned int  count,
                                            unsigned char *data) {

   unsigned int  dataSize;     // Bytes to program this row
   unsigned char rc;
   unsigned int  temp;
            int  doBlock;

   // Have to program within a row in each iteration
   // so there are size and alignment issues.
   while (count > 0) {
      // Bytes to end of current row
      dataSize = ICP_JM60_ROWSIZE - (addr & (ICP_JM60_ROWSIZE-1));
      if (dataSize > count)  // Take down to number actually remaining
         dataSize = count;

      // Check if empty block (all 0xFF)
      doBlock = FALSE;
      for (temp = 0; temp<dataSize; temp++)
         doBlock = doBlock || (data[temp] != 0xFF);

      if (doBlock) { // Only program non-Blank blocks
         print("Programming block %4.4X-%4.4X\r\n", addr, addr+dataSize-1);
         rc = bdm_usb_raw_send_ep0(ICP_PROGRAM_ROW,
                                   addr&0xFFFF, // LSB address
                                   addr>>16,    // MSB address
                                   dataSize,
                                   data);
         if (rc != ICP_RC_OK)
            return rc;

     //    rc = prefix(icp_get_result)();

         if (rc != ICP_RC_OK)
            return rc;
      }
      else
         print("Skipping block    %4.4X-%4.4X\r\n", addr, addr+dataSize-1);

      data  += dataSize;   // update location in buffer
      addr  += dataSize;   // update address
      count -= dataSize;   // update count
   }
   return ICP_RC_OK;
}

//! ICP mode - verify BDM Flash memory
//!
//! @param addr    32-bit memory address
//! @param count   number of bytes to verify
//! @param data    Pointer to buffer containing data
//!
USBDM_API unsigned int prefix(icp_verify)( unsigned int  addr,
                                           unsigned int  count,
                                           unsigned char *data) {
   unsigned       int  dataSize;
   unsigned       char rc;

   // Can verify MaxDataSize bytes at a time
   // Doesn't have alignment issues
   while (count > 0) {
      dataSize = count;
      if (dataSize > ICP_JM60_ROWSIZE)
         dataSize = ICP_JM60_ROWSIZE;

      print("Verifying block %4.4X-%4.4X\r\n", addr, addr+dataSize-1);
      rc = bdm_usb_raw_send_ep0(ICP_VERIFY_ROW,
                                addr&0xFFFF, // LSB address
                                addr>>16,    // MSB address
                                dataSize,
                                data);
      if (rc != ICP_RC_OK)
         return rc;

  //   rc = prefix(icp_get_result)();

      if (rc != ICP_RC_OK)
         return rc;

      data  += dataSize;   // update location in buffer
      addr  += dataSize;   // update address
      count -= dataSize;   // update count
   }
   return ICP_RC_OK;
}

//! ICP mode - erase BDM Flash memory
//!
//! @param addr    32-bit memory address
//! @param count   number of bytes to erase
//!
USBDM_API unsigned int prefix(icp_erase)( unsigned int addr,
                                          unsigned int count) {
   unsigned int blockSize;
   unsigned char rc;

   // Flash memory is organised in blocks
   // Only an entire block can be erased at a time
   while (count > 0) {
      blockSize = count;
      if (blockSize > ICP_JM60_BLOCKSIZE)
         blockSize = ICP_JM60_BLOCKSIZE;

      print("Erasing block %4.4X-%4.4X\r\n", addr, addr+blockSize-1);
      rc = bdm_usb_raw_send_ep0(ICP_ERASE_PAGE,
                                addr&0xFFFF, // LSB address
                                addr>>16,    // MSB address
                                0,
                                NULL);
      if (rc != ICP_RC_OK)
         return rc;

      rc = prefix(icp_get_result)();

      if (rc != ICP_RC_OK)
         return rc;

      addr  += blockSize;   // update address
      count -= blockSize;   // update count
   }
   return ICP_RC_OK;
}

//! ICP mode - get result of last ICP command
//!
USBDM_API unsigned int prefix(icp_get_result)( void ) {

   unsigned char rc;
   unsigned char buff[10];

   rc = bdm_usb_raw_recv_ep0( ICP_GET_RESULT,
                              0,
                              0,
                              sizeof(buff),
                              buff);
   return buff[0];
}

#if defined (COMMANDLINE)
//! Execute debug command (various, see \ref DebugSubCommands)
//!
//! @return 16-bit value
//!
void bdm_debug(char *usb_data) {
   usb_data[0] = 7;  // receive up to 7 bytes
   usb_data[1] = CMD_DEBUG;
   bdm_usb_recv_ep0(usb_data);
}
#endif

//! Passes a set option command to the BDM
//!
//! @param option Option number
//! @param value  Value for the option
//!
unsigned int bdm_set_option(unsigned char option, unsigned int value) {
   int rc;

   usb_data[0] = 2;  // receive up to 2 bytes
   usb_data[1] = CMD_OPTION;
   usb_data[2] = option;
   usb_data[3] = (unsigned char) value&0xFF;
   usb_data[4] = (unsigned char) (value>>8)&0xFF;
   rc = bdm_usb_recv_ep0(usb_data);
   if (rc <= 0){
      return BDM_RC_USB_ERROR;
   }

   return rc;
}

//! Initialise DLL
//!
#if !defined(COMMANDLINE)
static
#endif
void dll_initialize(HINSTANCE _hDLLInst){

   hDLLInst = _hDLLInst;
   openLogFile(NULL);
   bdm_usb_init();       /* initialise USB */
}

//! Un-initialise DLL
//!
#if !defined(COMMANDLINE)
static
#endif
void dll_uninitialize(void) {
   closeLogFile();
   hDLLInst = NULL;
}

#if defined(DLL)

 BOOL WINAPI DllMain(HINSTANCE _hDLLInst,
                     DWORD fdwReason,
                     LPVOID lpvReserved) {

   switch (fdwReason) {
      case DLL_PROCESS_ATTACH:
         dll_initialize(_hDLLInst);
         break;
      case DLL_PROCESS_DETACH:
         dll_uninitialize();
         break;
      case DLL_THREAD_ATTACH:
         //dll_initialize(_hDLLInst);
         break;
      case DLL_THREAD_DETACH:
         //dll_uninitialize();
         break;
 }
return TRUE;
}

#endif
