/*! \file
    \brief Provides Debug logging facilities

    \verbatim
    Turbo BDM Light - message log
    Copyright (C) 2005  Daniel Malik

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    \endverbatim
*/
#include <windows.h>
#include <stdio.h>
#include <time.h>
#include "log.h"
#include "version.h"

#ifdef TBLCF
#define DEF_NAME "tblcf"
#else
#define DEF_NAME "usbdm"
#endif

#ifdef LOG
   //! File handle for logging file
   static FILE *log_file=NULL;
#endif

#ifdef  LOG
	/*! \brief Provides a print function which prints data into a log file.
	 *
	 *  @param format Format and parameters as for printf()
	 */
	void print(const char *format, ...) {
		va_list list;

		if (log_file == NULL)
		   return;

		va_start(list, format);
		//fprintf(log_file, "%04.3f: ",1.0*GetTickCount()/1000);
		vfprintf(log_file, format, list);
		va_end(list);
		fflush(log_file);
	}

	/*! \brief Print a formatted dump of binary data in Hex
	 *
	 * @param data Pointer to data to print
	 * @param size Number of bytes to print
	 */
	void printDump(unsigned char *data, unsigned int size) {
		unsigned int i=0;

		if (log_file == NULL)
         return;

		while(size--) {
			fprintf(log_file,"%02X ",*(data++));
			if (((++i)%16)==0) fprintf(log_file,"\r\n");
		}
		if (((i)%16)!=0) fprintf(log_file,"\r\n");
		fflush(log_file);
	}

	/*! \brief Open the log file
	 *
	 *  @param LogFileName Path of file to open for logging. \n
	 *                     If NULL a default path is used.
	 */
	void openLogFile(char *LogFileName){
	   char path[MAX_PATH];
	   time_t time_now;
#if 0
      char *charp;

      if (LogFileName == NULL) {
         GetModuleFileName(NULL,path,sizeof(path));
         charp=strchr(path,'\\');
         if (charp!=NULL) {
            while (strchr(charp+1,'\\'))
               charp=strchr(charp+1,'\\'); /* find the last backslash */
            strcpy(charp+1, DEF_NAME "_dll.log");
            //log_file=fopen(path,"wb"); // Disable for now
         }
	   }
#endif

      if (log_file == NULL) {
         strcpy(path, "C:\\" DEF_NAME "_dll.log");
         log_file=fopen(path,"wb");
      }
      print("Log file path: %s\r\n",path);
      print(DEF_NAME ".dll v%1d.%1d. Compiled on %s, %s.\r\n",
            USBDM_DLL_VERSION/16,USBDM_DLL_VERSION&0x0f,__DATE__,__TIME__);
      time(&time_now);
      print("Log file created on: %s\r\n", ctime(&time_now));
	}

   /*!  \brief Close the log file
    *
    */
   void closeLogFile(void){
       time_t time_now;

      if (log_file == NULL)
         return;

       time(&time_now);
       print("End of log file: %s\r", ctime(&time_now));
       fclose(log_file);
	}
#else
	void print(const char *format, ...) {
		return;
	}

	void printDump(unsigned char *data, unsigned int size) {
		return;
	}

   void openLogFile(char *LogFileName){
      return;
   }

   void closeLogFile(void){
      return;
   }
#endif

